# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2018 The Google AI Language Team Authors and
# The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import collections

from nemo import logging


def convert_examples_to_features(
    examples, tokenizer, max_seq_length, doc_stride, max_query_length, has_groundtruth,
):
    """Loads a data file into a list of `InputBatch`s."""

    unique_id = 1000000000

    features = []
    for (example_index, example) in enumerate(examples):
        query_tokens = tokenizer.text_to_tokens(example.question_text)

        if len(query_tokens) > max_query_length:
            query_tokens = query_tokens[0:max_query_length]

        # context: index of token -> index of word
        tok_to_orig_index = []
        # context: index of word -> index of first token in token list
        orig_to_tok_index = []
        # context without white spaces after tokenization
        all_doc_tokens = []
        # doc tokens is word separated context
        for (i, token) in enumerate(example.doc_tokens):
            orig_to_tok_index.append(len(all_doc_tokens))
            sub_tokens = tokenizer.text_to_tokens(token)
            for sub_token in sub_tokens:
                tok_to_orig_index.append(i)
                all_doc_tokens.append(sub_token)

        # idx of query token start and end in context
        tok_start_position = None
        tok_end_position = None
        if has_groundtruth and example.is_impossible:
            tok_start_position = -1
            tok_end_position = -1
        if has_groundtruth and not example.is_impossible:
            tok_start_position = orig_to_tok_index[example.start_position]
            if example.end_position < len(example.doc_tokens) - 1:
                tok_end_position = orig_to_tok_index[example.end_position + 1] - 1
            else:
                tok_end_position = len(all_doc_tokens) - 1

            (tok_start_position, tok_end_position) = _improve_answer_span(
                all_doc_tokens, tok_start_position, tok_end_position, tokenizer, example.answer_text
            )

        # The -3 accounts for tokenizer.cls_token, tokenizer.sep_token and tokenizer.eos_token
        # doc_spans contains all possible contexts options of given length
        max_tokens_for_doc = max_seq_length - len(query_tokens) - 3
        _DocSpan = collections.namedtuple("DocSpan", ["start", "length"])
        doc_spans = []
        start_offset = 0
        while start_offset < len(all_doc_tokens):
            length = len(all_doc_tokens) - start_offset
            if length > max_tokens_for_doc:
                length = max_tokens_for_doc
            doc_spans.append(_DocSpan(start=start_offset, length=length))
            if start_offset + length == len(all_doc_tokens):
                break
            start_offset += min(length, doc_stride)

        for (doc_span_index, doc_span) in enumerate(doc_spans):
            tokens = []
            # maps context tokens idx in final input -> word idx in context
            token_to_orig_map = {}
            token_is_max_context = {}
            segment_ids = []
            tokens.append(tokenizer.bos_token)
            segment_ids.append(0)
            for token in query_tokens:
                tokens.append(token)
                segment_ids.append(0)
            tokens.append(tokenizer.sep_token)
            segment_ids.append(0)

            for i in range(doc_span.length):
                split_token_index = doc_span.start + i
                token_to_orig_map[len(tokens)] = tok_to_orig_index[split_token_index]

                is_max_context = _check_is_max_context(doc_spans, doc_span_index, split_token_index)
                token_is_max_context[len(tokens)] = is_max_context
                tokens.append(all_doc_tokens[split_token_index])
                segment_ids.append(1)
            tokens.append(tokenizer.eos_token)
            segment_ids.append(1)

            input_ids = tokenizer.tokens_to_ids(tokens)

            # The mask has 1 for real tokens and 0 for padding tokens.
            # Only real tokens are attended to.
            input_mask = [1] * len(input_ids)

            # Zero-pad up to the sequence length.
            while len(input_ids) < max_seq_length:
                input_ids.append(tokenizer.pad_id)
                input_mask.append(0)
                segment_ids.append(0)

            assert len(input_ids) == max_seq_length
            assert len(input_mask) == max_seq_length
            assert len(segment_ids) == max_seq_length

            # calculate start and end position in final array
            # of tokens in answer if no answer,
            # 0 for both pointing to tokenizer.cls_token
            start_position = None
            end_position = None
            if has_groundtruth and not example.is_impossible:
                doc_start = doc_span.start
                doc_end = doc_span.start + doc_span.length - 1
                out_of_span = False
                if not (tok_start_position >= doc_start and tok_end_position <= doc_end):
                    out_of_span = True
                if out_of_span:
                    start_position = 0
                    end_position = 0
                else:
                    doc_offset = len(query_tokens) + 2
                    start_position = tok_start_position - doc_start + doc_offset
                    end_position = tok_end_position - doc_start + doc_offset
            if has_groundtruth and example.is_impossible:
                # if our document chunk does not contain
                # an annotation we throw it out, since there is nothing
                # to predict.
                start_position = 0
                end_position = 0

            if example_index < 1:
                logging.info("*** Example ***")
                logging.info("unique_id: %s" % (unique_id))
                logging.info("example_index: %s" % (example_index))
                logging.info("doc_span_index: %s" % (doc_span_index))
                logging.info("tokens: %s" % " ".join(tokens))
                logging.info(
                    "token_to_orig_map: %s" % " ".join(["%d:%d" % (x, y) for (x, y) in token_to_orig_map.items()])
                )
                logging.info(
                    "token_is_max_context: %s"
                    % " ".join(["%d:%s" % (x, y) for (x, y) in token_is_max_context.items()])
                )
                logging.info("input_ids: %s" % " ".join([str(x) for x in input_ids]))
                logging.info("input_mask: %s" % " ".join([str(x) for x in input_mask]))
                logging.info("segment_ids: %s" % " ".join([str(x) for x in segment_ids]))
                if has_groundtruth and example.is_impossible:
                    logging.info("impossible example")
                if has_groundtruth and not example.is_impossible:
                    answer_text = " ".join(tokens[start_position : (end_position + 1)])
                    logging.info("start_position: %d" % (start_position))
                    logging.info("end_position: %d" % (end_position))
                    logging.info("answer: %s" % (answer_text))

            features.append(
                InputFeatures(
                    unique_id=unique_id,
                    example_index=example_index,
                    doc_span_index=doc_span_index,
                    tokens=tokens,
                    token_to_orig_map=token_to_orig_map,
                    token_is_max_context=token_is_max_context,
                    input_ids=input_ids,
                    input_mask=input_mask,
                    segment_ids=segment_ids,
                    start_position=start_position,
                    end_position=end_position,
                    is_impossible=example.is_impossible,
                )
            )
            unique_id += 1

    return features


def _improve_answer_span(doc_tokens, input_start, input_end, tokenizer, orig_answer_text):
    """Returns tokenized answer spans that
    better match the annotated answer."""
    tok_answer_text = " ".join(tokenizer.text_to_tokens(orig_answer_text))

    for new_start in range(input_start, input_end + 1):
        for new_end in range(input_end, new_start - 1, -1):
            text_span = " ".join(doc_tokens[new_start : (new_end + 1)])
            if text_span == tok_answer_text:
                return (new_start, new_end)

    return (input_start, input_end)


def _check_is_max_context(doc_spans, cur_span_index, position):
    """Check if this is the 'max context' doc span for the token.

    Because of the sliding window approach taken to scoring documents,
    a single token can appear in multiple documents.

    Example:
        Doc: the man went to the store and bought a gallon of milk
        Span A: the man went to the
        Span B: to the store and bought
        Span C: and bought a gallon of
        ...

    Now the word 'bought' will have two scores from spans B and C. We only
    want to consider the score with "maximum context", which we define as
    the *minimum* of its left and right context (the *sum* of left and
    right context will always be the same, of course).

    In the example the maximum context for 'bought' would be span C since
    it has 1 left context and 3 right context, while span B has 4 left context
    and 0 right context.

    Code adapted from the code by the Google AI and HuggingFace.
    """
    best_score = None
    best_span_index = None
    for (span_index, doc_span) in enumerate(doc_spans):
        end = doc_span.start + doc_span.length - 1
        if position < doc_span.start:
            continue
        if position > end:
            continue
        num_left_context = position - doc_span.start
        num_right_context = end - position
        score = min(num_left_context, num_right_context) + 0.01 * doc_span.length
        if best_score is None or score > best_score:
            best_score = score
            best_span_index = span_index

    return cur_span_index == best_span_index


class InputFeatures(object):
    """A single set of features of data."""

    def __init__(
        self,
        unique_id,
        example_index,
        doc_span_index,
        tokens,
        token_to_orig_map,
        token_is_max_context,
        input_ids,
        input_mask,
        segment_ids,
        start_position=None,
        end_position=None,
        is_impossible=None,
    ):
        self.unique_id = unique_id
        self.example_index = example_index
        self.doc_span_index = doc_span_index
        self.tokens = tokens
        self.token_to_orig_map = token_to_orig_map
        self.token_is_max_context = token_is_max_context
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids
        self.start_position = start_position
        self.end_position = end_position
        self.is_impossible = is_impossible



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
This file contains code artifacts adapted from the original implementation:
https://github.com/thu-coai/ConvLab-2/blob/master/convlab2/util/multiwoz/state.py
"""

__all__ = ['init_state']


def init_state():
    """
    Inits thate of the dialogue.
    Returns:
        4-tuple containing: Empty user utterance, empty system utterance, empty dialogue history
        and empty dialogue state.
    """
    # Create empty state - a dictionary with requires keys.
    # Keeping the original format to be "compatible" with ConvLab2.
    state = dict(user_action="", system_action=[], belief_state={}, request_state={}, terminated=False, history=[])

    # Create the MultiWOZ "belief state" structure.
    state['belief_state'] = {
        "police": {"book": {"booked": []}, "semi": {}},
        "hotel": {
            "book": {"booked": [], "people": "", "day": "", "stay": ""},
            "semi": {"name": "", "area": "", "parking": "", "pricerange": "", "stars": "", "internet": "", "type": ""},
        },
        "attraction": {"book": {"booked": []}, "semi": {"type": "", "name": "", "area": ""}},
        "restaurant": {
            "book": {"booked": [], "people": "", "day": "", "time": ""},
            "semi": {"food": "", "pricerange": "", "name": "", "area": "",},
        },
        "hospital": {"book": {"booked": []}, "semi": {"department": ""}},
        "taxi": {"book": {"booked": []}, "semi": {"leaveAt": "", "destination": "", "departure": "", "arriveBy": ""}},
        "train": {
            "book": {"booked": [], "people": ""},
            "semi": {"leaveAt": "", "destination": "", "day": "", "arriveBy": "", "departure": ""},
        },
    }
    return state['user_action'], state['system_action'], state['belief_state'], state['history']



# Copyright (c) 2019 NVIDIA Corporation
import collections
import json
import os
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from nemo.collections.asr.parts import manifest, parsers
from nemo.utils import logging


class _Collection(collections.UserList):
    """List of parsed and preprocessed data."""

    OUTPUT_TYPE = None  # Single element output type.


class Text(_Collection):
    """Simple list of preprocessed text entries, result in list of tokens."""

    OUTPUT_TYPE = collections.namedtuple('TextEntity', 'tokens')

    def __init__(self, texts: List[str], parser: parsers.CharParser):
        """Instantiates text manifest and do the preprocessing step.

        Args:
            texts: List of raw texts strings.
            parser: Instance of `CharParser` to convert string to tokens.
        """

        data, output_type = [], self.OUTPUT_TYPE
        for text in texts:
            tokens = parser(text)

            if tokens is None:
                logging.warning("Fail to parse '%s' text line.", text)
                continue

            data.append(output_type(tokens))

        super().__init__(data)


class FromFileText(Text):
    """Another form of texts manifest with reading from file."""

    def __init__(self, file: str, parser: parsers.CharParser):
        """Instantiates text manifest and do the preprocessing step.

        Args:
            file: File path to read from.
            parser: Instance of `CharParser` to convert string to tokens.
        """

        texts = self.__parse_texts(file)

        super().__init__(texts, parser)

    @staticmethod
    def __parse_texts(file: str) -> List[str]:
        if not os.path.exists(file):
            raise ValueError('Provided texts file does not exists!')

        _, ext = os.path.splitext(file)
        if ext == '.csv':
            texts = pd.read_csv(file)['transcript'].tolist()
        elif ext == '.json':  # Not really a correct json.
            texts = list(item['text'] for item in manifest.item_iter(file))
        else:
            with open(file, 'r') as f:
                texts = f.readlines()

        return texts


class AudioText(_Collection):
    """List of audio-transcript text correspondence with preprocessing."""

    OUTPUT_TYPE = collections.namedtuple(
        typename='AudioTextEntity', field_names='id audio_file duration text_tokens offset text_raw speaker',
    )

    def __init__(
        self,
        ids: List[int],
        audio_files: List[str],
        durations: List[float],
        texts: List[str],
        offsets: List[str],
        speakers: List[Optional[int]],
        parser: parsers.CharParser,
        min_duration: Optional[float] = None,
        max_duration: Optional[float] = None,
        max_number: Optional[int] = None,
        do_sort_by_duration: bool = False,
        index_by_file_id: bool = False,
    ):
        """Instantiates audio-text manifest with filters and preprocessing.

        Args:
            ids: List of examples positions.
            audio_files: List of audio files.
            durations: List of float durations.
            texts: List of raw text transcripts.
            offsets: List of duration offsets or None.
            speakers: List of optional speakers ids.
            parser: Instance of `CharParser` to convert string to tokens.
            min_duration: Minimum duration to keep entry with (default: None).
            max_duration: Maximum duration to keep entry with (default: None).
            max_number: Maximum number of samples to collect.
            do_sort_by_duration: True if sort samples list by duration. Not compatible with index_by_file_id.
            index_by_file_id: If True, saves a mapping from filename base (ID) to index in data.
        """

        output_type = self.OUTPUT_TYPE
        data, duration_filtered, num_filtered, total_duration = [], 0.0, 0, 0.0
        if index_by_file_id:
            self.mapping = {}

        for id_, audio_file, duration, offset, text, speaker in zip(
            ids, audio_files, durations, offsets, texts, speakers,
        ):
            # Duration filters.
            if min_duration is not None and duration < min_duration:
                duration_filtered += duration
                num_filtered += 1
                continue

            if max_duration is not None and duration > max_duration:
                duration_filtered += duration
                num_filtered += 1
                continue

            text_tokens = parser(text)
            if text_tokens is None:
                duration_filtered += duration
                num_filtered += 1
                continue

            total_duration += duration

            data.append(output_type(id_, audio_file, duration, text_tokens, offset, text, speaker))
            if index_by_file_id:
                file_id, _ = os.path.splitext(os.path.basename(audio_file))
                self.mapping[file_id] = len(data) - 1

            # Max number of entities filter.
            if len(data) == max_number:
                break

        if do_sort_by_duration:
            if index_by_file_id:
                logging.warning("Tried to sort dataset by duration, but cannot since index_by_file_id is set.")
            else:
                data.sort(key=lambda entity: entity.duration)

        logging.info("Dataset loaded with %d files totalling %.2f hours", len(data), total_duration / 3600)
        logging.info("%d files were filtered totalling %.2f hours", num_filtered, duration_filtered / 3600)

        super().__init__(data)


class ASRAudioText(AudioText):
    """`AudioText` collector from asr structured json files."""

    def __init__(self, manifests_files: Union[str, List[str]], *args, **kwargs):
        """Parse lists of audio files, durations and transcripts texts.

        Args:
            manifests_files: Either single string file or list of such -
                manifests to yield items from.
            *args: Args to pass to `AudioText` constructor.
            **kwargs: Kwargs to pass to `AudioText` constructor.
        """

        ids, audio_files, durations, texts, offsets, speakers = [], [], [], [], [], []
        for item in manifest.item_iter(manifests_files):
            ids.append(item['id'])
            audio_files.append(item['audio_file'])
            durations.append(item['duration'])
            texts.append(item['text'])
            offsets.append(item['offset'])
            speakers.append(item['speaker'])

        super().__init__(ids, audio_files, durations, texts, offsets, speakers, *args, **kwargs)


class SpeechLabel(_Collection):
    """List of audio-label correspondence with preprocessing."""

    OUTPUT_TYPE = collections.namedtuple(typename='SpeechLabelEntity', field_names='audio_file duration label offset',)

    def __init__(
        self,
        audio_files: List[str],
        durations: List[float],
        labels: List[Union[int, str]],
        offsets: List[Optional[float]],
        min_duration: Optional[float] = None,
        max_duration: Optional[float] = None,
        max_number: Optional[int] = None,
        do_sort_by_duration: bool = False,
    ):
        """Instantiates audio-label manifest with filters and preprocessing.

        Args:
            audio_files: List of audio files.
            durations: List of float durations.
            labels: List of labels.
            offsets: List of offsets or None.
            min_duration: Minimum duration to keep entry with (default: None).
            max_duration: Maximum duration to keep entry with (default: None).
            max_number: Maximum number of samples to collect.
            do_sort_by_duration: True if sort samples list by duration.
        """

        output_type = self.OUTPUT_TYPE
        data, duration_filtered = [], 0.0
        for audio_file, duration, command, offset in zip(audio_files, durations, labels, offsets):
            # Duration filters.
            if min_duration is not None and duration < min_duration:
                duration_filtered += duration
                continue

            if max_duration is not None and duration > max_duration:
                duration_filtered += duration
                continue

            data.append(output_type(audio_file, duration, command, offset))

            # Max number of entities filter.
            if len(data) == max_number:
                break

        if do_sort_by_duration:
            data.sort(key=lambda entity: entity.duration)

        logging.info(
            "Filtered duration for loading collection is %f.", duration_filtered,
        )
        self.uniq_labels = sorted(set(map(lambda x: x.label, data)))
        logging.info("# {} files loaded accounting to # {} labels".format(len(data), len(self.uniq_labels)))

        super().__init__(data)


class ASRSpeechLabel(SpeechLabel):
    """`SpeechLabel` collector from structured json files."""

    def __init__(self, manifests_files: Union[str, List[str]], *args, **kwargs):
        """Parse lists of audio files, durations and transcripts texts.

        Args:
            manifests_files: Either single string file or list of such -
                manifests to yield items from.
            *args: Args to pass to `SpeechLabel` constructor.
            **kwargs: Kwargs to pass to `SpeechLabel` constructor.
        """
        audio_files, durations, labels, offsets = [], [], [], []

        for item in manifest.item_iter(manifests_files, parse_func=self.__parse_item):
            audio_files.append(item['audio_file'])
            durations.append(item['duration'])
            labels.append(item['label'])
            offsets.append(item['offset'])

        super().__init__(audio_files, durations, labels, offsets, *args, **kwargs)

    def __parse_item(self, line: str, manifest_file: str) -> Dict[str, Any]:
        item = json.loads(line)

        # Audio file
        if 'audio_filename' in item:
            item['audio_file'] = item.pop('audio_filename')
        elif 'audio_filepath' in item:
            item['audio_file'] = item.pop('audio_filepath')
        else:
            raise ValueError(
                f"Manifest file has invalid json line " f"structure: {line} without proper audio file key."
            )
        item['audio_file'] = os.path.expanduser(item['audio_file'])

        # Duration.
        if 'duration' not in item:
            raise ValueError(f"Manifest file has invalid json line " f"structure: {line} without proper duration key.")

        # Label.
        if 'command' in item:
            item['label'] = item.pop('command')
        elif 'target' in item:
            item['label'] = item.pop('target')
        elif 'label' in item:
            pass
        else:
            raise ValueError(f"Manifest file has invalid json line " f"structure: {line} without proper label key.")

        item = dict(
            audio_file=item['audio_file'],
            duration=item['duration'],
            label=item['label'],
            offset=item.get('offset', None),
        )

        return item



# Copyright (c) 2019 NVIDIA Corporation
__all__ = [
    'WarmupPolicy',
    'WarmupHoldPolicy',
    'SquareAnnealing',
    'CosineAnnealing',
    'WarmupAnnealing',
    'InverseSquareRootAnnealing',
    'SquareRootAnnealing',
    'PolynomialDecayAnnealing',
    'PolynomialHoldDecayAnnealing',
]

import inspect
import math
import sys
from abc import ABC, abstractmethod


class _LRPolicy(ABC):
    """Base class for defining learning rate policies"""

    @abstractmethod
    def __call__(self, initial_lr, step, epoch):
        """Calculate and return learning rate for current epoch and step.

        Args:
            initial_lr: Initial learning rate to be passed to optimizer
            step: Current step number
            epoch: Current epoch number

        Returns:
            Learning rate value as single float number

        """

        pass


class WarmupPolicy(_LRPolicy):
    """Adds warmup kwargs and warmup logic to lr policy.

    All arguments should be passed as kwargs for clarity,

    Args:
        warmup_steps: Number of training steps in warmup stage
        warmup_ratio: Ratio of warmup steps to total steps
        total_steps: Total number of steps while training or `None` for
            infinite training

    """

    def __init__(self, *, warmup_steps=None, warmup_ratio=None, total_steps=None):
        assert not (
            warmup_steps is not None and warmup_ratio is not None
        ), "Either use particular number of step or ratio"
        assert warmup_ratio is None or total_steps is not None, "If there is a ratio, there should be a total steps"

        super().__init__()

        self.total_steps = total_steps
        if warmup_steps is not None:
            self.warmup_steps = warmup_steps
        elif warmup_ratio is not None:
            self.warmup_steps = int(warmup_ratio * total_steps)
        else:
            self.warmup_steps = 0

    def __call__(self, initial_lr, step, epoch):
        if step < self.warmup_steps:
            return initial_lr * (step + 1) / (self.warmup_steps + 1)
        if step > self.total_steps:
            return 0
        return self._get_lr(initial_lr, step, epoch)

    def _get_lr(self, initial_lr, step, epoch):
        """Simple const lr policy"""
        return initial_lr


class WarmupHoldPolicy(WarmupPolicy):
    """Variant of WarmupPolicy which maintains high learning rate for a defined number of steps.

    All arguments should be passed as kwargs for clarity,

    Args:
        warmup_steps: Number of training steps in warmup stage
        warmup_ratio: Ratio of warmup steps to total steps
        hold_steps: Number of training steps to hold the learning rate after warm up
        hold_ratio: Ratio of hold steps to total steps
        total_steps: Total number of steps while training or `None` for
            infinite training

    """

    def __init__(
        self, *, warmup_steps=None, warmup_ratio=None, hold_steps=None, hold_ratio=None, total_steps=None, min_lr=0.0
    ):
        assert not (hold_steps is not None and hold_ratio is not None), "Either use particular number of step or ratio"
        assert hold_ratio is None or total_steps is not None, "If there is a ratio, there should be a total steps"
        super().__init__(warmup_steps=warmup_steps, warmup_ratio=warmup_ratio, total_steps=total_steps)

        self._min_lr = min_lr
        self._last_warmup_lr = 0.0

        if hold_steps is not None:
            self.hold_steps = hold_steps + self.warmup_steps
        elif hold_ratio is not None:
            self.hold_steps = int(hold_ratio * total_steps) + self.warmup_steps
        else:
            self.hold_steps = 0

    def __call__(self, initial_lr, step, epoch):
        # Warmup phase
        if step < self.warmup_steps:
            warmup_lr = initial_lr * (step + 1) / (self.warmup_steps + 1)
            return warmup_lr

        # Hold phase
        if (step >= self.warmup_steps) and (step < self.hold_steps):
            return initial_lr

        if step > self.total_steps:
            return self._min_lr
        return self._get_lr(initial_lr, step, epoch)


def _squareroot_annealing(initial_lr, step, total_steps, min_lr):
    mult = ((total_steps - step) / total_steps) ** 0.5
    out_lr = initial_lr * mult
    out_lr = max(out_lr, min_lr)
    return out_lr


def _square_annealing(initial_lr, step, total_steps, min_lr):
    mult = ((total_steps - step) / total_steps) ** 2
    out_lr = initial_lr * mult
    out_lr = max(out_lr, min_lr)
    return out_lr


def _cosine_annealing(initial_lr, step, total_steps, min_lr):
    mult = 0.5 * (1 + math.cos(math.pi * step / total_steps))
    out_lr = (initial_lr - min_lr) * mult + min_lr
    return out_lr


def _poly_decay(initial_lr, step, decay_steps, power, min_lr, cycle):
    if cycle:
        multiplier = 1.0 if step == 0 else math.ceil(step / decay_steps)
        decay_steps *= multiplier
    else:
        step = min(step, decay_steps)
    p = step / decay_steps
    lr = (initial_lr - min_lr) * math.pow(1.0 - p, power)
    lr += min_lr
    return lr


class SquareAnnealing(WarmupPolicy):
    def __init__(self, total_steps, min_lr=1e-5, **kwargs):
        super().__init__(total_steps=total_steps, **kwargs)
        self.min_lr = min_lr

    def _get_lr(self, initial_lr, step, epoch):
        return _square_annealing(
            initial_lr=initial_lr,
            step=step - self.warmup_steps,
            total_steps=self.total_steps - self.warmup_steps,
            min_lr=self.min_lr,
        )


class SquareRootAnnealing(WarmupPolicy):
    def __init__(self, total_steps, min_lr=0, **kwargs):
        super().__init__(total_steps=total_steps, **kwargs)
        self.min_lr = min_lr

    def _get_lr(self, initial_lr, step, epoch):
        return _squareroot_annealing(
            initial_lr=initial_lr, step=step, total_steps=self.total_steps, min_lr=self.min_lr,
        )


class CosineAnnealing(WarmupPolicy):
    def __init__(self, total_steps, min_lr=0, **kwargs):
        self.min_lr = min_lr
        super().__init__(total_steps=total_steps, **kwargs)

    def _get_lr(self, initial_lr, step, epoch):
        if initial_lr < self.min_lr:
            raise ValueError(
                f"{self} received an initial learning rate that " f"was lower than the minimum learning rate."
            )
        return _cosine_annealing(
            initial_lr=initial_lr,
            step=step - self.warmup_steps,
            total_steps=self.total_steps - self.warmup_steps,
            min_lr=self.min_lr,
        )


class WarmupAnnealing(WarmupPolicy):
    def __init__(self, total_steps, **kwargs):
        super().__init__(total_steps=total_steps, **kwargs)

    def _get_lr(self, initial_lr, step, epoch):
        progress = float(step / self.total_steps)
        warmup_ratio = float(self.warmup_steps / self.total_steps)

        mult = max((progress - 1.0) / (warmup_ratio - 1.0), 0.0)
        out_lr = initial_lr * mult

        return out_lr


class InverseSquareRootAnnealing(WarmupPolicy):
    def __init__(self, total_steps, **kwargs):
        super().__init__(total_steps=total_steps, **kwargs)

    def _get_lr(self, initial_lr, step, epoch):
        denom = ((step + 1) / (self.warmup_steps + 1)) ** 0.5
        out_lr = initial_lr / denom
        return out_lr


class PolynomialDecayAnnealing(WarmupPolicy):
    def __init__(self, total_steps, min_lr=0.0, power=1.0, cycle=False, **kwargs):
        super().__init__(total_steps=total_steps, **kwargs)
        self.min_lr = min_lr
        self.power = power
        self.cycle = cycle

    def _get_lr(self, initial_lr, step, epoch):
        return _poly_decay(
            initial_lr,
            step=step - self.warmup_steps,
            decay_steps=self.total_steps - self.warmup_steps,
            power=self.power,
            min_lr=self.min_lr,
            cycle=self.cycle,
        )


class PolynomialHoldDecayAnnealing(WarmupHoldPolicy):
    def __init__(self, total_steps, min_lr=0.0, power=1.0, cycle=False, **kwargs):
        super().__init__(total_steps=total_steps, **kwargs)
        self.min_lr = min_lr
        self.power = power
        self.cycle = cycle

    def _get_lr(self, initial_lr, step, epoch):
        return _poly_decay(
            initial_lr,
            step=step - self.hold_steps,
            decay_steps=self.total_steps - max(self.warmup_steps, self.hold_steps),
            power=self.power,
            min_lr=self.min_lr,
            cycle=self.cycle,
        )


def get_all_lr_classes():
    """ Get all LR classes defined within this module
    """
    lr_classes = {}
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj) and name != 'ABC':
            lr_classes[name] = obj
    return lr_classes


def get_lr_policy(lr_policy, **kwargs):
    lr_classes = get_all_lr_classes()
    if lr_policy not in lr_classes:
        raise ValueError(
            f'{lr_policy} is not a supported lr policy. ' f'Supported lr policies are {lr_classes.keys()}.'
        )
    return lr_classes[lr_policy](**kwargs)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Utility functions for Token Classification NLP tasks
Some parts of this code were adapted from the HuggingFace library at
https://github.com/huggingface/pytorch-pretrained-BERT
"""

import os
import random

import h5py
import numpy as np
from torch.utils.data import Dataset

from nemo import logging
from nemo.collections.nlp.data.datasets.datasets_utils.data_preprocessing import get_stats
from nemo.collections.nlp.utils.callback_utils import list2str

__all__ = ['BertTextClassificationDataset']


class BertTextClassificationDataset(Dataset):
    """A dataset class that converts from raw data to
    a dataset that can be used by DataLayerNM.

    Args:
        input_file (str): file to sequence + label.
            the first line is header (sentence [tab] label)
            each line should be [sentence][tab][label]
        max_seq_length (int): max sequence length minus 2 for [CLS] and [SEP]
        tokenizer (Tokenizer): such as NemoBertTokenizer
        num_samples (int): number of samples you want to use for the dataset.
            If -1, use all dataset. Useful for testing.
    """

    def __init__(
        self, input_file, max_seq_length, tokenizer, num_samples=-1, shuffle=False, use_cache=False,
    ):
        self.input_file = input_file
        self.max_seq_length = max_seq_length
        self.tokenizer = tokenizer
        self.num_samples = num_samples
        self.use_cache = use_cache
        self.shuffle = shuffle
        self.vocab_size = self.tokenizer.tokenizer.vocab_size

        if use_cache:
            data_dir, filename = os.path.split(input_file)
            vocab_size = getattr(tokenizer, "vocab_size", 0)
            tokenizer_type = type(tokenizer.tokenizer).__name__
            cached_features_file = os.path.join(
                data_dir,
                "cached_{}_{}_{}_{}".format(
                    filename[:-4], tokenizer_type, str(max_seq_length), str(vocab_size), '.hdf5'
                ),
            )

        if use_cache and os.path.exists(cached_features_file):
            self.load_cached_features(cached_features_file)
        else:
            with open(input_file, "r") as f:
                sent_labels, all_sent_subtokens = [], []
                sent_lengths = []
                too_long_count = 0

                lines = f.readlines()[1:]
                logging.info(f'{input_file}: {len(lines)}')

                if shuffle or num_samples > -1:
                    random.seed(0)
                    random.shuffle(lines)
                    if num_samples > 0:
                        lines = lines[:num_samples]

                for index, line in enumerate(lines):
                    if index % 20000 == 0:
                        logging.debug(f"Processing line {index}/{len(lines)}")

                    line_splited = line.strip().split()
                    sent_label = int(line_splited[-1])
                    sent_labels.append(sent_label)
                    sent_words = line_splited[:-1]
                    sent_subtokens = [tokenizer.cls_token]

                    for word in sent_words:
                        word_tokens = tokenizer.text_to_tokens(word)
                        sent_subtokens.extend(word_tokens)

                    sent_subtokens.append(tokenizer.sep_token)

                    all_sent_subtokens.append(sent_subtokens)
                    sent_lengths.append(len(sent_subtokens))
            get_stats(sent_lengths)

            for i in range(len(all_sent_subtokens)):
                if len(all_sent_subtokens[i]) > max_seq_length:
                    shorten_sent = all_sent_subtokens[i][-max_seq_length + 1 :]
                    all_sent_subtokens[i] = [tokenizer.cls_token] + shorten_sent
                    too_long_count += 1

            logging.info(
                f'{too_long_count} out of {len(sent_lengths)} \
                        sentences with more than {max_seq_length} subtokens.'
            )

            self.convert_sequences_to_features(all_sent_subtokens, sent_labels, tokenizer, max_seq_length)

            if self.use_cache:
                self.cache_features(cached_features_file, self.features)

                # update self.features to use features from hdf5
                self.load_cached_features(cached_features_file)

    def __len__(self):
        if self.use_cache:
            return len(self.features[0])

        else:
            return len(self.features)

    def __getitem__(self, idx):
        if self.use_cache:
            return (self.features[0][idx], self.features[1][idx], self.features[2][idx], self.features[3][idx])

        else:
            feature = self.features[idx]

            return (
                np.array(feature.input_ids),
                np.array(feature.segment_ids),
                np.array(feature.input_mask, dtype=np.long),
                feature.sent_label,
            )

    def convert_sequences_to_features(self, all_sent_subtokens, sent_labels, tokenizer, max_seq_length):
        """Loads a data file into a list of `InputBatch`s.
        """

        self.features = []
        for sent_id in range(len(all_sent_subtokens)):
            sent_subtokens = all_sent_subtokens[sent_id]
            sent_label = sent_labels[sent_id]

            input_ids = [tokenizer.tokens_to_ids(t) for t in sent_subtokens]

            # The mask has 1 for real tokens and 0 for padding tokens.
            # Only real tokens are attended to.
            input_mask = [1] * len(input_ids)

            # Zero-pad up to the sequence length.
            while len(input_ids) < max_seq_length:
                input_ids.append(0)
                input_mask.append(0)
            segment_ids = [0] * max_seq_length

            assert len(input_ids) == max_seq_length
            assert len(input_mask) == max_seq_length

            if sent_id < 5:
                logging.info("*** Example ***")
                logging.info("example_index: %s" % sent_id)
                logging.info("subtokens: %s" % " ".join(sent_subtokens))
                logging.info("sent_label: %s" % sent_label)
                logging.info("input_ids: %s" % list2str(input_ids))
                logging.info("input_mask: %s" % list2str(input_mask))

            self.features.append(
                InputFeatures(
                    sent_id=sent_id,
                    sent_label=sent_label,
                    input_ids=input_ids,
                    input_mask=input_mask,
                    segment_ids=segment_ids,
                )
            )

    def cache_features(self, cached_features_file, features):
        len_features = len(features)
        input_ids_array = np.zeros((len_features, self.max_seq_length))
        segment_ids_array = np.zeros((len_features, self.max_seq_length))
        input_mask_array = np.zeros((len_features, self.max_seq_length))
        sent_labels_array = np.zeros((len_features,))

        for idx in range(len_features):
            input_ids_array[idx] = features[idx].input_ids
            segment_ids_array[idx] = features[idx].segment_ids
            input_mask_array[idx] = features[idx].input_mask
            sent_labels_array[idx] = features[idx].sent_label

        f = h5py.File(cached_features_file, mode='w')
        f.create_dataset('input_ids', data=input_ids_array)
        f.create_dataset('segment_ids', data=segment_ids_array)
        f.create_dataset('input_mask', data=input_mask_array)
        f.create_dataset('sent_labels', data=sent_labels_array)
        f.close()

    def load_cached_features(self, cached_features_file):
        f = h5py.File(cached_features_file, 'r')
        keys = ['input_ids', 'segment_ids', 'input_mask', 'sent_labels']
        self.features = [np.asarray(f[key], dtype=np.long) for key in keys]
        f.close()
        logging.info(f'features restored from {cached_features_file}')

        if self.shuffle:
            np.random.seed(0)
            idx = np.arange(len(self))
            np.random.shuffle(idx)  # shuffle idx in place
            shuffled_features = [arr[idx] for arr in self.features]
            self.features = shuffled_features

        if self.num_samples > 0:
            truncated_features = [arr[0 : self.num_samples] for arr in self.features]
            self.features = truncated_features


class InputFeatures(object):
    """A single set of features of data."""

    def __init__(self, sent_id, sent_label, input_ids, input_mask, segment_ids):
        self.sent_id = sent_id
        self.sent_label = sent_label
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.nlp.nm.data_layers.text_datalayer import TextDataLayer
from nemo.core import ChannelType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['BertInferDataLayer']


class BertInferDataLayer(TextDataLayer):
    """
    Data layer to run infernce with BERT (get final hidden layer).

    Args:
        tokenizer (TokenizerSpec): tokenizer
        dataset (str): directory or a single file with dataset documents
        max_seq_length (int): maximum allowed length of the text segments
        mask_probability (float): probability of masking input sequence tokens
        batch_size (int): batch size in segments
        short_seeq_prob (float): Probability of creating sequences which are
            shorter than the maximum length.
            Defualts to 0.1.
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        input_ids: indices of tokens which constitute batches of text segments
            0: AxisType(BatchTag)

            1: AxisType(TimeTag)

        input_type_ids: indices of token types (e.g., sentences A & B in BERT)
            0: AxisType(BatchTag)

            1: AxisType(TimeTag)

        input_mask: bool tensor with 0s in place of tokens to be masked
            0: AxisType(BatchTag)

            1: AxisType(TimeTag)

        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_type_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_mask": NeuralType(('B', 'T'), ChannelType()),
        }

    def __init__(self, dataset_type, dataset_params, batch_size=1, shuffle=False):

        super().__init__(dataset_type, dataset_params, batch_size=batch_size, shuffle=shuffle)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import random

import torch
import torch.nn.functional as pt_f
from torch import nn

from nemo.backends.pytorch.common.parts import Attention
from nemo.backends.pytorch.nm import TrainableNM
from nemo.core import *
from nemo.utils.decorators import add_port_docs
from nemo.utils.misc import pad_to

__all__ = ['DecoderRNN', 'EncoderRNN']


class DecoderRNN(TrainableNM):
    """Simple RNN-based decoder with attention.

    Args:
        voc_size (int): Total number of symbols to use
        bos_id (int): Label position of start of string symbol
        hidden_size (int): Size of hidden vector to use in RNN
        attention_method (str): Method of using attention to pass in
            `Attention` constructor.
            Defaults to 'general'.
        attention_type (str): String type of attention describing time to apply
            attention. Could be on of ['post', 'none'].
            Defaults to 'post'.
        in_dropout (float): Float value of embedding dropout.
            Defaults to 0.2.
        gru_dropout (float): Float value of RNN interlayers dropout
            Defaults to 0.2.
        attn_dropout (float): Float value of attention dropout to pass to
            `Attention` constructor
            Defaults to 0.0.
        teacher_forcing (float): Probability of applying full teacher forcing
            method at each step.
            Defaults to 1.0.
        curriculum_learning (float): If teacher forcing is not applying, this
            value indicates probability of using target token from next step.
            Defaults to 0.5.
        rnn_type (str): Type of RNN to use. Could be one of ['gru', 'lstm'].
            Defaults to 'gru'.
        n_layers (int): Number of layers to use in RNN.
            Defaults to 2.
        tie_emb_out_weights (bool): Whether to tie embedding and output
            weights.
            Defaults to True.

    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            'targets': NeuralType(('B', 'T'), LabelsType()),
            'encoder_outputs': NeuralType(('B', 'T', 'D'), ChannelType(), True),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            'log_probs': NeuralType(('B', 'T', 'D'), LogprobsType()),
            'attention_weights': NeuralType(('B', 'T', 'T'), ChannelType(), True),
        }

    def __init__(
        self,
        voc_size,
        bos_id,
        hidden_size,
        attention_method='general',
        attention_type='post',
        in_dropout=0.2,
        gru_dropout=0.2,
        attn_dropout=0.0,
        teacher_forcing=1.0,
        curriculum_learning=0.5,
        rnn_type='gru',
        n_layers=2,
        tie_emb_out_weights=True,
    ):
        super().__init__()

        self.bos_id = bos_id
        self.attention_type = attention_type
        self.teacher_forcing = teacher_forcing
        self.curriculum_learning = curriculum_learning
        self.rnn_type = rnn_type

        voc_size = pad_to(voc_size, 8)  # 8-divisors trick
        self.embedding = nn.Embedding(voc_size, hidden_size)
        # noinspection PyTypeChecker
        self.in_dropout = nn.Dropout(in_dropout)
        rnn_class = getattr(nn, rnn_type.upper())
        self.rnn = rnn_class(
            hidden_size, hidden_size, n_layers, dropout=(0 if n_layers == 1 else gru_dropout), batch_first=True,
        )
        self.out = nn.Linear(hidden_size, voc_size)
        if tie_emb_out_weights:
            self.out.weight = nn.Parameter(self.embedding.weight)  # Weight tying
        self.attention = Attention(hidden_size, attention_method, dropout=attn_dropout)

        # self.apply(init_weights)
        # self.gru.apply(init_weights)
        self.to(self._device)

    def forward(self, targets, encoder_outputs=None):
        if (not self.training) or (random.random() <= self.teacher_forcing):  # Fast option
            # Removing last char (dont need to calculate loss) and add bos
            # noinspection PyTypeChecker
            decoder_inputs = pt_f.pad(targets[:, :-1], (1, 0), value=self.bos_id)  # BT
            log_probs, _, attention_weights = self.forward_step(decoder_inputs, encoder_outputs)
        else:
            log_probs, attention_weights = self.forward_cl(targets, encoder_outputs)

        return log_probs, attention_weights

    def forward_step(self, decoder_inputs, encoder_outputs=None, decoder_hidden=None):
        """(BT, BTC@?, hBC@?) -> (BTC, hBC, BTT@?)"""

        # Inputs
        decoder_inputs = self.embedding(decoder_inputs)
        # noinspection PyCallingNonCallable
        decoder_inputs = self.in_dropout(decoder_inputs)

        # RNN
        if self.rnn_type == 'gru' and decoder_hidden is not None:
            decoder_hidden = decoder_hidden[0]
        decoder_outputs, decoder_hidden = self.rnn(decoder_inputs, decoder_hidden)
        if self.rnn_type == 'gru':
            decoder_hidden = (decoder_hidden,)

        # Outputs
        attention_weights = None
        if self.attention_type == 'post':
            decoder_outputs, attention_weights = self.attention(decoder_outputs, encoder_outputs)
        decoder_outputs = self.out(decoder_outputs)

        # Log probs
        log_probs = pt_f.log_softmax(decoder_outputs, dim=-1)

        return log_probs, decoder_hidden, attention_weights

    def forward_cl(self, targets, encoder_outputs=None):
        """(BT, BTC@?) -> (BTC, BTT@?)"""

        decoder_input = torch.empty(targets.size(0), 1, dtype=torch.long, device=self._device).fill_(self.bos_id)
        decoder_hidden = None
        log_probs = []
        attention_weights = []

        max_len = targets.size(1)
        rands = torch.rand(max_len)  # Precalculate randomness
        for i in range(max_len):
            (step_log_prob, decoder_hidden, step_attention_weights,) = self.forward_step(
                decoder_input, encoder_outputs, decoder_hidden
            )
            log_probs.append(step_log_prob)
            attention_weights.append(step_attention_weights)

            if rands[i] <= self.curriculum_learning:
                decoder_input = targets[:, i].view(-1, 1).detach()
            else:
                decoder_input = step_log_prob.argmax(-1).detach()

        log_probs = torch.cat(log_probs, dim=1)
        if len(attention_weights) and attention_weights[0] is not None:
            attention_weights = torch.cat(attention_weights, dim=1)
        else:
            attention_weights = None

        return log_probs, attention_weights


class EncoderRNN(TrainableNM):
    """ Simple RNN-based encoder using GRU cells """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            'inputs': NeuralType(('B', 'T'), ChannelType()),
            'input_lens': NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            'outputs': NeuralType(('B', 'T', 'D'), ChannelType()),
            'hidden': NeuralType(('B', 'T', 'D'), ChannelType()),
        }

    def __init__(
        self, input_dim, emb_dim, hid_dim, dropout, n_layers=1, pad_idx=1, embedding_to_load=None, sum_hidden=True
    ):
        super().__init__()
        self.dropout = nn.Dropout(dropout)
        self.embedding = nn.Embedding(input_dim, emb_dim, padding_idx=pad_idx)
        if embedding_to_load is not None:
            self.embedding.weight.data.copy_(embedding_to_load)
        else:
            self.embedding.weight.data.normal_(0, 0.1)
        self.rnn = nn.GRU(emb_dim, hid_dim, n_layers, batch_first=True, dropout=dropout, bidirectional=True)
        self.sum_hidden = sum_hidden
        self.to(self._device)

    def forward(self, inputs, input_lens=None):
        embedded = self.embedding(inputs)
        embedded = self.dropout(embedded)
        if input_lens is not None:
            embedded = nn.utils.rnn.pack_padded_sequence(embedded, input_lens, batch_first=True)

        outputs, hidden = self.rnn(embedded)
        # outputs of shape (seq_len, batch, num_directions * hidden_size)
        # hidden of shape (num_layers * num_directions, batch, hidden_size)
        if input_lens is not None:
            outputs, _ = nn.utils.rnn.pad_packed_sequence(outputs, batch_first=True)
        else:
            outputs = outputs.transpose(0, 1)
        # outputs of shape: (batch, seq_len, num_directions * hidden_size)

        batch_size = hidden.size()[1]

        # separate final hidden states by layer and direction
        hidden = hidden.view(self.rnn.num_layers, 2 if self.rnn.bidirectional else 1, batch_size, self.rnn.hidden_size)
        hidden = hidden.transpose(2, 0).transpose(1, 2)
        # hidden shape: batch x num_layer x num_directions x hidden_size
        if self.sum_hidden and self.rnn.bidirectional:
            hidden = hidden[:, :, 0, :] + hidden[:, :, 1, :]
            outputs = outputs[:, :, : self.rnn.hidden_size] + outputs[:, :, self.rnn.hidden_size :]
        else:
            hidden = hidden.reshape(batch_size, self.rnn.num_layers, -1)
        # hidden is now of shape (batch, num_layer, [num_directions] * hidden_size)

        return outputs, hidden



# Copyright (c) 2019 NVIDIA Corporation
from statistics import mean

import torch

from nemo import logging


def compute_accuracy(tensors):
    logging.info(f"Train Loss: {str(tensors[0].item())}")
    output = tensors[1]
    target = tensors[2]
    res = []
    topk = (1,)
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        for k in topk:
            correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
    return "Train Top@1 accuracy  {0} ".format(res[0].item())


def eval_iter_callback(tensors, global_vars):
    if "eval_loss" not in global_vars.keys():
        global_vars["eval_loss"] = []
    for kv, v in tensors.items():
        if kv.startswith("loss"):
            global_vars["eval_loss"].append(torch.mean(torch.stack(v)).item())
            # global_vars['eval_loss'].append(v.item())

    if "top1" not in global_vars.keys():
        global_vars["top1"] = []

    output = None
    labels = None
    for kv, v in tensors.items():
        if kv.startswith("output"):
            # output = tensors[kv]
            output = torch.cat(tensors[kv])
        if kv.startswith("label"):
            # labels = tensors[kv]
            labels = torch.cat(tensors[kv])

    if output is None:
        raise Exception("output is None")

    res = []
    topk = (1,)
    with torch.no_grad():
        maxk = max(topk)
        batch_size = labels.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(labels.view(1, -1).expand_as(pred))

        for k in topk:
            correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))

    global_vars["top1"].append(res[0].item())


def eval_epochs_done_callback(global_vars):
    eloss = mean(global_vars["eval_loss"])
    etop1 = mean(global_vars["top1"])
    logging.info("Evaluation Loss: {0}".format(eloss))
    logging.info("Evaluation Top@1: {0}".format(etop1))
    for k in global_vars.keys():
        global_vars[k] = []
    return dict({"Evaluation Loss": eloss, "Evaluation Top@1": etop1})



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
This file contains code artifacts adapted from the original implementation:
https://github.com/google-research/google-research/blob/master/schema_guided_dst
"""

import collections
import os

import numpy as np
import torch

from nemo import logging
from nemo.collections.nlp.data.datasets.sgd_dataset import schema
from nemo.collections.nlp.data.datasets.sgd_dataset.schema_embedding_dataset import SchemaEmbeddingDataset
from nemo.collections.nlp.nm.data_layers.bert_inference_datalayer import BertInferDataLayer
from nemo.collections.nlp.utils.data_utils import concatenate

__all__ = ['SchemaPreprocessor']


class SchemaPreprocessor:
    """ 
    Convert the raw data to the standard format supported by
    StateTrackingSGDData.
    
    Args:
        data_dir (str) - Directory for the downloaded DSTC8/SGD data, which contains
            the dialogue files and schema files of all datasets (eg train, dev)
        dialogues_example_dir (str) - Directory where preprocessed DSTC8/SGD dialogues are stored
        schema_embedding_dir (str) - Directory where .npy file for embedding of
            entities (slots, values, intents) in the dataset_split's
            schema are stored.
        task_name (str) - The name of the task to train
        vocab_file (str) - The path to BERT vocab file
        do_lower_case - (bool) - Whether to lower case the input text.
            Should be True for uncased models and False for cased models.
        max_seq_length (int) - The maximum total input sequence length after
            WordPiece tokenization. Sequences longer than this will be
            truncated, and sequences shorter than this will be padded."
        tokenizer - tokenizer
        bert_model - pretrained BERT model
        dataset_split (str) - Dataset split for training / prediction (train/dev/test)
        overwrite_dial_file (bool) - Whether to generate a new file saving
            the dialogue examples overwrite_schema_emb_file,
        bert_ckpt_dir (str) - Directory containing pre-trained BERT checkpoint
        nf - NeuralModuleFactory
        mode(str): Schema embeddings initialization mode, baseline is ['CLS'] token embeddings
        from the last BERT layer
    """

    def __init__(
        self,
        data_dir,
        schema_embedding_dir,
        schema_config,
        tokenizer,
        bert_model,
        overwrite_schema_emb_files,
        bert_ckpt_dir,
        nf,
        add_carry_value,
        add_carry_status,
        datasets=['train', 'test', 'dev'],
        mode='baseline',
        is_trainable=False,
    ):

        # Dimension of the embedding for intents, slots and categorical slot values in
        # Maximum allowed number of categorical trackable slots for a service.
        self.schema_config = schema_config.copy()

        self.is_trainable = is_trainable
        self.datasets = datasets

        self._add_carry_value = add_carry_value
        self._add_carry_status = add_carry_status
        if self._add_carry_status:
            self._slot_status_size = 4
        else:
            self._slot_status_size = 3

        for dataset_split in ['train', 'test', 'dev']:
            if dataset_split not in self.datasets:
                logging.warning(
                    'WARNING: %s set was not included and won\'t be processed. Services from this dataset split '
                    + 'won\'t be supported',
                    dataset_split,
                )
        os.makedirs(schema_embedding_dir, exist_ok=True)

        tokenizer_type = type(tokenizer.tokenizer).__name__
        vocab_size = getattr(tokenizer, "vocab_size", 0)
        self.schema_embedding_file = os.path.join(
            schema_embedding_dir,
            "{}_{}_{}_{}_pretrained_schema_embedding.npy".format(
                '_'.join(self.datasets), mode, tokenizer_type, vocab_size
            ),
        )
        all_schema_json_paths = []
        for dataset_split in self.datasets:
            all_schema_json_paths.append(os.path.join(data_dir, dataset_split, "schema.json"))
        self.schemas = schema.Schema(
            all_schema_json_paths, add_carry_value=self._add_carry_value, add_carry_status=self._add_carry_status
        )

        if not os.path.exists(self.schema_embedding_file) or overwrite_schema_emb_files:
            # Generate the schema embeddings if needed or specified
            logging.info(f"Start generating the schema embeddings.")
            dataset_params = {
                "schema_config": schema_config,
                "tokenizer": tokenizer,
                "schemas": self.schemas,
            }
            emb_datalayer = BertInferDataLayer(
                dataset_type=SchemaEmbeddingDataset, dataset_params=dataset_params, batch_size=1, shuffle=False,
            )

            input_ids, input_mask, input_type_ids = emb_datalayer()

            hidden_states = bert_model(input_ids=input_ids, token_type_ids=input_type_ids, attention_mask=input_mask)
            evaluated_tensors = nf.infer(tensors=[hidden_states], checkpoint_dir=bert_ckpt_dir)

            master_device = not torch.distributed.is_initialized() or torch.distributed.get_rank() == 0
            if master_device:
                hidden_states = [concatenate(tensors) for tensors in evaluated_tensors]
                emb_datalayer.dataset.save_embeddings(hidden_states, self.schema_embedding_file, mode)
                logging.info(f"Finish generating the schema embeddings.")

        # wait until the master process writes to the schema embedding file
        if torch.distributed.is_initialized():
            torch.distributed.barrier()

        with open(self.schema_embedding_file, "rb") as f:
            self.schema_embeddings = np.load(f, allow_pickle=True)
            f.close()

    def get_schema_embeddings(self):
        # Convert from list of dict to dict of list
        schema_data_dict = collections.defaultdict(list)
        for service in self.schema_embeddings:
            schema_data_dict["cat_slot_emb"].append(service["cat_slot_emb"])
            schema_data_dict["cat_slot_value_emb"].append(service["cat_slot_value_emb"])
            schema_data_dict["noncat_slot_emb"].append(service["noncat_slot_emb"])
            schema_data_dict["req_slot_emb"].append(service["req_slot_emb"])
            schema_data_dict["intent_emb"].append(service["intent_emb"])
        return schema_data_dict

    def _get_schema_embedding_file_name(self):
        return self.schema_embedding_file

    def get_service_names_to_id_dict(self):
        return self.schemas._services_vocab

    def get_ids_to_service_names_dict(self):
        return self.schemas._services_id_to_vocab

    def update_slots_relation_list(self, slots_relation_list):
        self.schemas._slots_relation_list = slots_relation_list

    def get_slots_relation_list(self):
        return self.schemas._slots_relation_list



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.nlp.nm.trainables.common.common_utils import *
from nemo.collections.nlp.nm.trainables.common.encoder_rnn import *
from nemo.collections.nlp.nm.trainables.common.huggingface import *
from nemo.collections.nlp.nm.trainables.common.sequence_classification_nm import *
from nemo.collections.nlp.nm.trainables.common.sequence_regression_nm import *
from nemo.collections.nlp.nm.trainables.common.token_classification_nm import *
from nemo.collections.nlp.nm.trainables.common.transformer import *
from nemo.utils import logging

try:
    from nemo.collections.nlp.nm.trainables.common.megatron.megatron_utils import *

except Exception as e:
    logging.error('Failed to import Megatron utils: `{}` ({})'.format(str(e), type(e)))



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import os
import random

import h5py
import numpy as np
import torch
from torch.utils import data as pt_data

from nemo.backends.pytorch import DataLayerNM
from nemo.collections.nlp.data import BertPretrainingDataset, BertPretrainingPreprocessedDataset
from nemo.collections.nlp.nm.data_layers.text_datalayer import TextDataLayer
from nemo.core import ChannelType, LabelsType, MaskType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['BertPretrainingDataLayer', 'BertPretrainingPreprocessedDataLayer']


class BertPretrainingDataLayer(TextDataLayer):
    """
    Data layer for masked language modeling task for text data.

    Args:
        tokenizer (TokenizerSpec): tokenizer
        dataset (str): directory or a single file with dataset documents
        max_seq_length (int): maximum allowed length of the text segments
        mask_probability (float): probability of masking input sequence tokens
        batch_size (int): batch size in segments
        short_seeq_prob (float): Probability of creating sequences which are
            shorter than the maximum length.
            Defaults to 0.1.
        shuffle (bool): whether to shuffle data or not. Default: False.
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        input_ids:
            indices of tokens which constitute batches of masked text segments
        input_type_ids:
            tensor with 0's and 1's to denote the text segment type
        input_mask:
            bool tensor with 0s in place of tokens to be masked
        output_ids: indices of tokens which constitute batches of unmasked text segments
        output_mask: bool tensor with 0s in place of tokens to be masked
        labels: 0 or 1 for next sentence prediction classification
        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_type_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_mask": NeuralType(('B', 'T'), ChannelType()),
            "output_ids": NeuralType(('B', 'T'), LabelsType()),
            "output_mask": NeuralType(('B', 'T'), MaskType()),
            "labels": NeuralType(tuple('B'), LabelsType()),
        }

    def __init__(
        self, tokenizer, dataset, max_seq_length, mask_probability, short_seq_prob=0.1, batch_size=64, shuffle=False
    ):
        dataset_params = {
            'tokenizer': tokenizer,
            'dataset': dataset,
            'max_seq_length': max_seq_length,
            'mask_probability': mask_probability,
            'short_seq_prob': short_seq_prob,
        }
        super().__init__(BertPretrainingDataset, dataset_params, batch_size, shuffle=shuffle)


class BertPretrainingPreprocessedDataLayer(DataLayerNM):
    """
    Data layer for masked language modeling task for preprocessed data.

    Args:
        dataset (str): directory or a single file with dataset documents
        max_seq_length (int): maximum allowed length of the text segments
        batch_size (int): batch size in segments
        mode (str): model execution mode, e.g. "training"
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        input_ids:
            indices of tokens which constitute batches of masked text segments
        input_type_ids:
            tensor with 0's and 1's to denote the text segment type
        input_mask:
            bool tensor with 0s in place of tokens to be masked
        output_ids: indices of tokens which constitute batches of unmasked text segments
        output_mask: bool tensor with 0s in place of tokens to be masked
        labels: 0 or 1 for next sentence prediction classification
        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_type_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_mask": NeuralType(('B', 'T'), ChannelType()),
            "output_ids": NeuralType(('B', 'T'), LabelsType()),
            "output_mask": NeuralType(('B', 'T'), MaskType()),
            "labels": NeuralType(tuple('B'), LabelsType()),
        }

    def __init__(
        self, dataset, max_pred_length, mode, batch_size=64,
    ):
        super().__init__()
        if os.path.isdir(dataset):
            self.files = [
                os.path.join(dataset, f) for f in os.listdir(dataset) if os.path.isfile(os.path.join(dataset, f))
            ]
        else:
            self.files = [dataset]
        self.files.sort()
        self.num_files = len(self.files)
        self._batch_size = batch_size
        self.max_pred_length = max_pred_length
        self.mode = mode
        total_length = 0
        for f in self.files:
            fp = h5py.File(f, 'r')
            total_length += len(fp['input_ids'])
            fp.close()
        self.total_length = total_length

    def _collate_fn(self, x):
        num_components = len(x[0])
        components = [[] for _ in range(num_components)]
        batch_size = len(x)
        for i in range(batch_size):
            for j in range(num_components):
                components[j].append(x[i][j])
        src_ids, src_segment_ids, src_mask, tgt_ids, tgt_mask, sent_ids = [np.stack(x, axis=0) for x in components]
        src_ids = torch.Tensor(src_ids).long().to(self._device)
        src_segment_ids = torch.Tensor(src_segment_ids).long().to(self._device)
        src_mask = torch.Tensor(src_mask).long().to(self._device)
        tgt_ids = torch.Tensor(tgt_ids).long().to(self._device)
        tgt_mask = torch.Tensor(tgt_mask).long().to(self._device)
        sent_ids = torch.Tensor(sent_ids).long().to(self._device)
        return src_ids, src_segment_ids, src_mask, tgt_ids, tgt_mask, sent_ids

    def __len__(self):
        return self.total_length

    @property
    def dataset(self):
        return None

    @property
    def data_iterator(self):
        while True:
            if self.mode == "train":
                random.shuffle(self.files)
            for f_id in range(self.num_files):
                data_file = self.files[f_id]
                train_data = BertPretrainingPreprocessedDataset(
                    input_file=data_file, max_pred_length=self.max_pred_length
                )
                train_sampler = pt_data.RandomSampler(train_data)
                train_dataloader = pt_data.DataLoader(
                    dataset=train_data,
                    batch_size=self._batch_size,
                    collate_fn=self._collate_fn,
                    shuffle=False,
                    sampler=train_sampler,
                )
                for x in train_dataloader:
                    yield x
            if self.mode != "train":
                break



# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import argparse

from torch import optim

import nemo.utils.argparse as nm_argparse
from nemo.collections.cv.modules.data_layers import MNISTDataLayer
from nemo.collections.cv.modules.losses import NLLLoss
from nemo.collections.cv.modules.trainables import LeNet5
from nemo.core import DeviceType, NeuralGraph, NeuralModuleFactory, OperationMode
from nemo.utils import logging

if __name__ == "__main__":
    # Create the default parser.
    parser = argparse.ArgumentParser(parents=[nm_argparse.NemoArgParser()], conflict_handler='resolve')
    # Parse the arguments
    args = parser.parse_args()

    # Instantiate Neural Factory.
    nf = NeuralModuleFactory(local_rank=args.local_rank)

    # Data layers for training and validation.
    dl = MNISTDataLayer(height=32, width=32, train=True)
    dl_e = MNISTDataLayer(height=32, width=32, train=False)
    # The "model".
    lenet5 = LeNet5()
    # Loss.
    nll_loss = NLLLoss()

    # Create a training graph.
    with NeuralGraph(operation_mode=OperationMode.training) as training_graph:
        _, x, y, _ = dl()
        p = lenet5(images=x)
        loss = nll_loss(predictions=p, targets=y)

    # Display the graph summmary.
    logging.info(training_graph.summary())

    # Create a validation graph, starting from the second data layer.
    with NeuralGraph(operation_mode=OperationMode.evaluation) as evaluation_graph:
        _, x, y, _ = dl_e()
        p = lenet5(images=x)
        loss_e = nll_loss(predictions=p, targets=y)

    # Perform operations on GPU.
    training_graph.to(DeviceType.GPU)
    evaluation_graph.to(DeviceType.GPU)

    # Create optimizer.
    opt = optim.Adam(training_graph.parameters(), lr=0.001)

    # Print frequency.
    freq = 10
    # Train for 5 epochs.
    for epoch in range(5):
        # Configure data loader - once per epoch.
        # Just change the batch_size and turn sample shuffling on.
        training_graph.configure_data_loader(batch_size=128, shuffle=True)

        # Iterate over the whole dataset - in batches.
        for step, batch in enumerate(training_graph.get_batch()):

            # Reset the gradients.
            opt.zero_grad()

            # Forward pass.
            outputs = training_graph.forward(batch)
            # Print loss.
            if step % freq == 0:
                logging.info("Epoch: {} Step: {} Training Loss: {}".format(epoch, step, outputs.loss))

            # Backpropagate the gradients.
            training_graph.backward()

            # Update the parameters.
            opt.step()
        # Epoch ended.

        # Evaluate graph on test set.
        # Configure data loader - once per epoch.
        eval_losses = []
        evaluation_graph.configure_data_loader(batch_size=128)
        # Iterate over the whole dataset - in batches.
        for step, batch in enumerate(evaluation_graph.get_batch()):
            # Forward pass.
            outputs = evaluation_graph.forward(batch)
            eval_losses.append(outputs.loss)
        # Print avg. loss.
        logging.info("Epoch: {} Avg. Evaluation Loss: {}".format(epoch, sum(eval_losses) / len(eval_losses)))



# Source: https://github.com/tartley/colorama/blob/master/colorama/ansi.py
# Copyright: Jonathan Hartley 2013. BSD 3-Clause license.
#
# Copyright (C) NVIDIA CORPORATION. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.****

CSI = "\033["
OSC = "\033]"
BEL = "\007"


def code_to_chars(code):
    return CSI + str(code) + "m"


def set_title(title):
    return OSC + "2;" + title + BEL


def clear_screen(mode=2):
    return CSI + str(mode) + "J"


def clear_line(mode=2):
    return CSI + str(mode) + "K"


class AnsiCodes(object):
    def __init__(self):
        # the subclasses declare class attributes which are numbers.
        # Upon instantiation we define instance attributes, which are the same
        # as the class attributes but wrapped with the ANSI escape sequence
        for name in dir(self):
            if not name.startswith("_"):
                value = getattr(self, name)
                setattr(self, name, code_to_chars(value))


class AnsiCursor(object):
    def UP(self, n=1):
        return CSI + str(n) + "A"

    def DOWN(self, n=1):
        return CSI + str(n) + "B"

    def FORWARD(self, n=1):
        return CSI + str(n) + "C"

    def BACK(self, n=1):
        return CSI + str(n) + "D"

    def POS(self, x=1, y=1):
        return CSI + str(y) + ";" + str(x) + "H"


class AnsiFore(AnsiCodes):
    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37
    RESET = 39

    # These are fairly well supported, but not part of the standard.
    LIGHTBLACK_EX = 90
    LIGHTRED_EX = 91
    LIGHTGREEN_EX = 92
    LIGHTYELLOW_EX = 93
    LIGHTBLUE_EX = 94
    LIGHTMAGENTA_EX = 95
    LIGHTCYAN_EX = 96
    LIGHTWHITE_EX = 97


class AnsiBack(AnsiCodes):
    BLACK = 40
    RED = 41
    GREEN = 42
    YELLOW = 43
    BLUE = 44
    MAGENTA = 45
    CYAN = 46
    WHITE = 47
    RESET = 49

    # These are fairly well supported, but not part of the standard.
    LIGHTBLACK_EX = 100
    LIGHTRED_EX = 101
    LIGHTGREEN_EX = 102
    LIGHTYELLOW_EX = 103
    LIGHTBLUE_EX = 104
    LIGHTMAGENTA_EX = 105
    LIGHTCYAN_EX = 106
    LIGHTWHITE_EX = 107


class AnsiStyle(AnsiCodes):
    BRIGHT = 1
    DIM = 2
    NORMAL = 22
    RESET_ALL = 0


Fore = AnsiFore()
Back = AnsiBack()
Style = AnsiStyle()
Cursor = AnsiCursor()



# -*- coding: utf-8 -*-
# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
# Copyright (C) IBM Corporation 2019
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

__author__ = "Younes Bouhadjar, Vincent Marois, Tomasz Kornuta"

"""
This file contains code artifacts adapted from the original implementation:
https://github.com/IBM/pytorchpipe/blob/develop/ptp/components/models/vision/convnet_encoder.py
"""


from typing import Optional

import numpy as np
import torch.nn as nn

from nemo.backends.pytorch.nm import TrainableNM
from nemo.core.neural_types import AxisKind, AxisType, ImageFeatureValue, ImageValue, NeuralType
from nemo.utils import logging
from nemo.utils.decorators import add_port_docs

__all__ = ['ConvNetEncoder']


class ConvNetEncoder(TrainableNM):
    """
    A simple image encoder consisting of 3 consecutive convolutional layers.
    The parameters of input image (height, width and depth) are not hardcoded
    so the encoder can be adjusted for a given application (image dimensions).
    """

    def __init__(
        self,
        input_depth: int,
        input_height: int,
        input_width: int,
        conv1_out_channels: int = 64,
        conv1_kernel_size: int = 3,
        conv1_stride: int = 1,
        conv1_padding: int = 0,
        maxpool1_kernel_size: int = 2,
        conv2_out_channels: int = 32,
        conv2_kernel_size: int = 3,
        conv2_stride: int = 1,
        conv2_padding: int = 0,
        maxpool2_kernel_size: int = 2,
        conv3_out_channels: int = 16,
        conv3_kernel_size: int = 3,
        conv3_stride: int = 1,
        conv3_padding: int = 0,
        maxpool3_kernel_size: int = 2,
        name: Optional[str] = None,
    ):
        """
        Constructor of the a simple CNN.

        The overall structure of this CNN is as follows:

            (Conv1 -> MaxPool1 -> ReLu) -> (Conv2 -> MaxPool2 -> ReLu) -> (Conv3 -> MaxPool3 -> ReLu)

        The parameters that the user can change are:

         - For Conv1, Conv2 & Conv3: number of output channels, kernel size, stride and padding.
         - For MaxPool1, MaxPool2 & MaxPool3: Kernel size


        .. note::

            We are using the default values of ``dilatation``, ``groups`` & ``bias`` for ``nn.Conv2D``.

            Similarly for the ``stride``, ``padding``, ``dilatation``, ``return_indices`` & ``ceil_mode`` of \
            ``nn.MaxPool2D``.

        Args: 
            input_depth: Depth of the input image
            input_height: Height of the input image
            input_width: Width of the input image
            convX_out_channels: Number of output channels of layer X (X=1,2,3)
            convX_kernel_size: Kernel size of layer X (X=1,2,3)
            convX_stride: Stride of layer X (X=1,2,3)
            convX_padding: Padding of layer X (X=1,2,3)
            name: Name of the module (DEFAULT: None)
        """
        # Call base constructor.
        TrainableNM.__init__(self, name=name)

        # Get input image information from the global parameters.
        self._input_depth = input_depth
        self._input_height = input_height
        self._input_width = input_width

        # Retrieve the Conv1 parameters.
        self._conv1_out_channels = conv1_out_channels
        self._conv1_kernel_size = conv1_kernel_size
        self._conv1_stride = conv1_stride
        self._conv1_padding = conv1_padding

        # Retrieve the MaxPool1 parameter.
        self._maxpool1_kernel_size = maxpool1_kernel_size

        # Retrieve the Conv2 parameters.
        self._conv2_out_channels = conv2_out_channels
        self._conv2_kernel_size = conv2_kernel_size
        self._conv2_stride = conv2_stride
        self._conv2_padding = conv2_padding

        # Retrieve the MaxPool2 parameter.
        self._maxpool2_kernel_size = maxpool2_kernel_size

        # Retrieve the Conv3 parameters.
        self._conv3_out_channels = conv3_out_channels
        self._conv3_kernel_size = conv3_kernel_size
        self._conv3_stride = conv3_stride
        self._conv3_padding = conv3_padding

        # Retrieve the MaxPool3 parameter.
        self._maxpool3_kernel_size = maxpool3_kernel_size

        # We can compute the spatial size of the output volume as a function of the input volume size (W),
        # the receptive field size of the Conv Layer neurons (F), the stride with which they are applied (S),
        # and the amount of zero padding used (P) on the border.
        # The corresponding equation is conv_size = ((W−F+2P)/S)+1.

        # doc for nn.Conv2D: https://pytorch.org/docs/stable/nn.html#torch.nn.Conv2d
        # doc for nn.MaxPool2D: https://pytorch.org/docs/stable/nn.html#torch.nn.MaxPool2d

        # ----------------------------------------------------
        # Conv1
        self._conv1 = nn.Conv2d(
            in_channels=self._input_depth,
            out_channels=self._conv1_out_channels,
            kernel_size=self._conv1_kernel_size,
            stride=self._conv1_stride,
            padding=self._conv1_padding,
            dilation=1,
            groups=1,
            bias=True,
        )

        width_features_conv1 = np.floor(
            ((self._input_width - self._conv1_kernel_size + 2 * self._conv1_padding) / self._conv1_stride) + 1
        )
        height_features_conv1 = np.floor(
            ((self._input_height - self._conv1_kernel_size + 2 * self._conv1_padding) / self._conv1_stride) + 1
        )

        # ----------------------------------------------------
        # MaxPool1
        self._maxpool1 = nn.MaxPool2d(kernel_size=self._maxpool1_kernel_size)

        width_features_maxpool1 = np.floor(
            ((width_features_conv1 - self._maxpool1_kernel_size + 2 * self._maxpool1.padding) / self._maxpool1.stride)
            + 1
        )

        height_features_maxpool1 = np.floor(
            ((height_features_conv1 - self._maxpool1_kernel_size + 2 * self._maxpool1.padding) / self._maxpool1.stride)
            + 1
        )

        # ----------------------------------------------------
        # Conv2
        self._conv2 = nn.Conv2d(
            in_channels=self._conv1_out_channels,
            out_channels=self._conv2_out_channels,
            kernel_size=self._conv2_kernel_size,
            stride=self._conv2_stride,
            padding=self._conv2_padding,
            dilation=1,
            groups=1,
            bias=True,
        )

        width_features_conv2 = np.floor(
            ((width_features_maxpool1 - self._conv2_kernel_size + 2 * self._conv2_padding) / self._conv2_stride) + 1
        )
        height_features_conv2 = np.floor(
            ((height_features_maxpool1 - self._conv2_kernel_size + 2 * self._conv2_padding) / self._conv2_stride) + 1
        )

        # ----------------------------------------------------
        # MaxPool2
        self._maxpool2 = nn.MaxPool2d(kernel_size=self._maxpool2_kernel_size)

        width_features_maxpool2 = np.floor(
            ((width_features_conv2 - self._maxpool2_kernel_size + 2 * self._maxpool2.padding) / self._maxpool2.stride)
            + 1
        )
        height_features_maxpool2 = np.floor(
            ((height_features_conv2 - self._maxpool2_kernel_size + 2 * self._maxpool2.padding) / self._maxpool2.stride)
            + 1
        )

        # ----------------------------------------------------
        # Conv3
        self._conv3 = nn.Conv2d(
            in_channels=self._conv2_out_channels,
            out_channels=self._conv3_out_channels,
            kernel_size=self._conv3_kernel_size,
            stride=self._conv3_stride,
            padding=self._conv3_padding,
            dilation=1,
            groups=1,
            bias=True,
        )

        width_features_conv3 = np.floor(
            ((width_features_maxpool2 - self._conv3_kernel_size + 2 * self._conv3_padding) / self._conv3_stride) + 1
        )
        height_features_conv3 = np.floor(
            ((height_features_maxpool2 - self._conv3_kernel_size + 2 * self._conv3_padding) / self._conv3_stride) + 1
        )

        # ----------------------------------------------------
        # MaxPool3
        self._maxpool3 = nn.MaxPool2d(kernel_size=self._maxpool3_kernel_size)

        width_features_maxpool3 = np.floor(
            ((width_features_conv3 - self._maxpool3_kernel_size + 2 * self._maxpool3.padding) / self._maxpool3.stride)
            + 1
        )

        height_features_maxpool3 = np.floor(
            ((height_features_conv3 - self._maxpool1_kernel_size + 2 * self._maxpool3.padding) / self._maxpool3.stride)
            + 1
        )

        # Rememvber the output dims.
        self._feature_map_height = height_features_maxpool3
        self._feature_map_width = width_features_maxpool3
        self._feature_map_depth = self._conv3_out_channels

        # Log info about dimensions.
        logging.info('Input shape: [-1, {}, {}, {}]'.format(self._input_depth, self._input_height, self._input_width))
        logging.debug('Computed output shape of each layer:')
        logging.debug(
            '  * Conv1: [-1, {}, {}, {}]'.format(self._conv1_out_channels, height_features_conv1, width_features_conv1)
        )
        logging.debug(
            '  * MaxPool1: [-1, {}, {}, {}]'.format(
                self._conv1_out_channels, height_features_maxpool1, width_features_maxpool1
            )
        )
        logging.debug(
            '  * Conv2: [-1, {}, {}, {}]'.format(self._conv2_out_channels, height_features_conv2, width_features_conv2)
        )
        logging.debug(
            '  * MaxPool2: [-1, {}, {}, {}]'.format(
                self._conv2_out_channels, height_features_maxpool2, width_features_maxpool2
            )
        )
        logging.debug(
            '  * Conv3: [-1, {}, {}, {}]'.format(
                self._conv3_out_channels, height_features_conv3, width_features_conv3,
            )
        )
        logging.debug(
            '  * MaxPool3: [-1, {}, {}, {}]'.format(
                self._conv3_out_channels, width_features_maxpool3, height_features_maxpool3
            )
        )
        logging.info(
            'Output shape: [-1, {}, {}, {}]'.format(
                self._feature_map_depth, self._feature_map_height, self._feature_map_width
            )
        )

    @property
    @add_port_docs()
    def input_ports(self):
        """
        Returns definitions of module input ports.
        """
        return {
            "inputs": NeuralType(
                axes=(
                    AxisType(kind=AxisKind.Batch),
                    AxisType(kind=AxisKind.Channel, size=self._input_depth),
                    AxisType(kind=AxisKind.Height, size=self._input_height),
                    AxisType(kind=AxisKind.Width, size=self._input_width),
                ),
                elements_type=ImageValue(),
            )
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """
        Returns definitions of module output ports.
        """
        return {
            "outputs": NeuralType(
                axes=(
                    AxisType(kind=AxisKind.Batch),
                    AxisType(kind=AxisKind.Channel, size=self._feature_map_depth),
                    AxisType(kind=AxisKind.Height, size=self._feature_map_height),
                    AxisType(kind=AxisKind.Width, size=self._feature_map_width),
                ),
                elements_type=ImageFeatureValue(),
            )
        }

    def forward(self, inputs):
        """
        Forward pass of the convnet module.

        :param data_streams: DataStreams({'inputs','outputs'}), where:

        Args:
            inputs: Batch of inputs to be processed [BATCH_SIZE x INPUT_DEPTH x INPUT_HEIGHT x INPUT_WIDTH]
        
        Returns:
            Batch of outputs [BATCH_SIZE x OUTPUT_DEPTH x OUTPUT_HEIGHT x OUTPUT_WIDTH]

        """
        # apply Convolutional layer 1
        out_conv1 = self._conv1(inputs)

        # apply max_pooling and relu
        out_maxpool1 = nn.functional.relu(self._maxpool1(out_conv1))

        # apply Convolutional layer 2
        out_conv2 = self._conv2(out_maxpool1)

        # apply max_pooling and relu
        out_maxpool2 = nn.functional.relu(self._maxpool2(out_conv2))

        # apply Convolutional layer 3
        out_conv3 = self._conv3(out_maxpool2)

        # apply max_pooling and relu
        out_maxpool3 = nn.functional.relu(self._maxpool3(out_conv3))

        # Return output.
        return out_maxpool3



# -*- coding: utf-8 -*-

# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from collections.abc import MutableMapping
from typing import Dict, List, Optional, Union

from frozendict import frozendict

from nemo.utils import logging
from nemo.utils.neural_graph.connection import StepModulePort


class GraphInput(object):
    """ A helper class represenging a single bound input. """

    def __init__(self, ntype: "NeuralType"):
        """ 
        Initializes object.

        Args:
            ntype: a NeuralType object.
        """
        # (Neural) Type of input.
        self._ntype = ntype
        # List of StepModulePort tuples to which this input links to (step number, module name, port name).
        self._consumers = []

    def bind(self, step_module_ports: Union[StepModulePort, List[StepModulePort]]):
        """ Binds the (step-module-ports) to this "graph input".
            Add "consumers" of this graph input (modules attached to this port),
            so when one actually will pass the NmTensor, those modules will be connected.

            Args:
                step_module_ports: A single StepModulePort OR a list of StepModulePort tuples to be added.
        """
        # Handle both single port and lists of ports to be bound.
        if type(step_module_ports) is not list:
            step_module_ports = [step_module_ports]
        # Interate through "consumers" on the list and add them to bound input.
        for smp in step_module_ports:
            self._consumers.append(smp)

    @property
    def ntype(self) -> "NeuralType":
        """
            Returns:
                NeuralType of a given input.
        """
        return self._ntype

    @property
    def consumers(self) -> List[StepModulePort]:
        """ 
            Returns:
                List of bound modules i.e. (step number, module name, port name) tupes.
        """
        return self._consumers


class GraphInputs(MutableMapping):
    '''
        A specialized dictionary that contains bound inputs of a Neural Graph.
    '''

    def __init__(self):
        """
            Initializes an empty dictionary.
        """
        self._inputs = {}

    def __setitem__(self, key: str, value: Union["NeuralType", GraphInput]):
        """
            This method is used to "create" a bound input, i.e. copy definition from indicated module input port.

            Args:
                key: name of the input port of the Neural Graph.
                value: NeuralType (or GraphInput) that will be set.
            
            Raises:
                KeyError: Definition of a previously bound port is not allowed.
                TypeError: Port definition must be must be a NeuralType or GraphInput type.
        """
        # Make sure that a proper object was passed here.
        if type(value).__name__ == "NeuralType":
            ntype = value
        elif isinstance(value, GraphInput):
            ntype = value.ntype
        else:
            raise TypeError("Port `{}` definition must be must be a NeuralType or GraphInput type".format(key))

        if key in self._inputs.keys():
            if self._inputs[key].ntype == ntype:
                raise KeyError("Overwriting definition of a previously bound port `{}` is not allowed".format(key))
            # Else: do nothing.
        else:
            # Ok, add definition to list of mapped (module, port)s.
            # Note: for now, there are no mapped modules, so copy only the (neural) type.
            self._inputs[key] = GraphInput(ntype=ntype)

    def __getitem__(self, key: str) -> GraphInput:
        """
            Returns the bound input associated with the given key.

            Args:
                key: Name of the bound input.
        """
        return self._inputs[key]

    def __delitem__(self, key: str):
        """
            Raises:
                TypeError as deletion of a bound input port is not allowed.
        """
        raise TypeError("Deletion of a bound input port is not allowed")

    def __iter__(self):
        """ 
            Returns:
                Iterator over the dict of bound inputs.
        """
        return iter(self._inputs)

    def __len__(self) -> int:
        """
            Return:
                The number of bound inputs.
        """
        return len(self._inputs)

    @property
    def definitions(self) -> Dict[str, "NeuralType"]:
        """
            Property returns definitions of the input ports by extracting them on the fly from list.

            ..info:
                This property actually returns a FrozenDict containing port definitions to indicate that
                port definitions SHOULD not be used during the actual binding.
            
            Returns:
                Dictionary of neural types associated with bound inputs.
        """
        # Extract port definitions (Neural Types) from the inputs list.
        return frozendict({k: v.ntype for k, v in self._inputs.items()})

    def has_binding(self, step_number: int, port_name: str) -> Optional[str]:
        """ 
            Checks if there is a binding leading to a given step number (module) and its given port. 
            (module name is redundant, thus skipped in this test).

            Returns:
                key in the list of the (bound) input ports that leads to a given step (module)/port
                or None if the binding was not found.
        """
        for key, binding in self._inputs.items():
            for (step, _, port) in binding.consumers:
                if step == step_number and port == port_name:
                    return key
        # Binding not found.
        return None

    def serialize(self) -> List[str]:
        """ Method responsible for serialization of the graph inputs.

            Returns:
                List containing mappings (input -> step.module.input_port).
        """
        serialized_inputs = []
        # Iterate through "bindings" (GraphInputs).
        for key, binding in self._inputs.items():
            # Get type.
            ntype_str = str(binding.ntype)
            for (step, module, port) in binding.consumers:
                # Serialize: input -> step.module.port | ntype
                target = str(step) + "." + module + "." + port
                # Serialize!
                serialized_inputs.append(key + "->" + target + " | " + ntype_str)
        # Return the result.
        return serialized_inputs

    @classmethod
    def deserialize(cls, serialized_inputs: List[str], modules: Dict[str, 'NeuralModule']):
        """ 
            Class method responsible for deserialization of graph inputs.

            Args:
                serialized_inputs: A list of serialized inputs in the form of ("input->module.input_port")
                modules: List of modules required for neural type copying/checking.

            Returns:
                Dictionary with deserialized inputs.
        """
        inputs = GraphInputs()
        # Iterate through serialized inputs one by one.
        for i in serialized_inputs:
            # Deserialize!
            [key, consumer_ntype] = i.split("->")
            [consumer, ntype_str] = consumer_ntype.split(" | ")
            [consumer_step, consumer_name, consumer_port_name] = consumer.split(".")
            # Add the input.
            if key not in inputs.keys():
                # Get neural type from module input port definition.
                ntype = modules[consumer_name].input_ports[consumer_port_name]
                # Make sure the graph bound  port type matches the deserialized type.
                assert ntype_str == str(ntype)

                # Create a new input.
                inputs[key] = ntype
            # Bind the "consumers".
            inputs[key].bind(StepModulePort(int(consumer_step), consumer_name, consumer_port_name))
        # Done.
        return inputs



import torch


def char_lm_metrics(
    chars_log_probs_batches, chars_targets_batches, targets_texts_batches, pad_id,
):
    """Calculate metrics for language modeling.

    Args:
        chars_log_probs_batches:
        chars_targets_batches:
        targets_texts_batches:
        pad_id:

    Returns:
        A tuple of two:
            bpc
            perplexity

    """

    bpcs, ppls = [], []
    for log_probs, targets, texts in zip(chars_log_probs_batches, chars_targets_batches, targets_texts_batches):
        target_log_probs = log_probs.gather(2, targets.unsqueeze(2)).squeeze(2)
        pad_mask = (targets != pad_id).long()
        nll = -(target_log_probs * pad_mask.float()).sum(-1)
        char_lens = pad_mask.float().sum(-1)
        word_lens = torch.tensor([len(text.split()) for text in texts], dtype=torch.float, device=char_lens.device,)
        bpc = nll / char_lens
        ppl = 2 ** (nll / word_lens)
        # ppl = 2 ** (bpc * ENG_MWN)  # ~5.3
        bpcs.append(bpc)
        ppls.append(ppl)

    bpc, ppl = torch.cat(bpcs), torch.cat(ppls)

    return bpc, ppl



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import re

from nemo import logging
from nemo.collections.nlp.data.tokenizers.tokenizer_spec import TokenizerSpec
from nemo.collections.nlp.data.tokenizers.tokenizer_utils import MODEL_SPECIAL_TOKENS, TOKENIZERS

__all__ = [
    'NemoBertTokenizer',
]


def handle_quotes(text):
    text_ = ""
    quote = 0
    i = 0
    while i < len(text):
        if text[i] == "\"":
            if quote % 2:
                text_ = text_[:-1] + "\""
            else:
                text_ += "\""
                i += 1
            quote += 1
        else:
            text_ += text[i]
        i += 1
    return text_


def remove_spaces(text):
    text = text.replace("( ", "(")
    text = text.replace(" )", ")")
    text = text.replace("[ ", "[")
    text = text.replace(" ]", "]")
    text = text.replace(" / ", "/")
    text = text.replace("„ ", "„")
    text = text.replace(" - ", "-")
    text = text.replace(" ' ", "'")
    text = re.sub(r'([0-9])( )([\.,])', '\\1\\3', text)
    text = re.sub(r'([\.,])( )([0-9])', '\\1\\3', text)
    text = re.sub(r'([0-9])(:)( )([0-9])', '\\1\\2\\4', text)
    text = text.replace(" %", "%")
    text = text.replace("$ ", "$")
    text = re.sub(r'([^0-9])(,)([0-9])', '\\1\\2 \\3', text)
    return text


class NemoBertTokenizer(TokenizerSpec):
    def __init__(self, pretrained_model=None, vocab_file=None, bert_derivative='bert', do_lower_case=False):
        '''
        The user needs to specify pretrained_model name or vocab file and bert_derivative

        pretrained_model (str):name of the pretrained model from the hugging face list,
            for example: bert-base-cased
            To see the list of pretrained models, call:
            huggingface_utils.get_bert_models_list()
        vocab_file: File containing the vocabulary.
        bert_derivative: for example: 'bert', 'roberta', 'albert'. Only used when vocab_file specified.
        '''
        if pretrained_model:
            bert_derivative = pretrained_model.split('-')[0]
            logging.info(f'Deriving bert model type from pretrained model name.')

        if bert_derivative in TOKENIZERS:
            tokenizer_cls = TOKENIZERS[bert_derivative]
        else:
            raise ValueError(
                "Bert_derivative value {bert_derivative} is not currently supported"
                + " Please choose from the following list: {TOKENIZERS.keys()}"
            )

        if pretrained_model:
            self.tokenizer = tokenizer_cls.from_pretrained(pretrained_model)
        elif vocab_file:
            self.tokenizer = tokenizer_cls(vocab_file=vocab_file, do_lower_case=do_lower_case)
        else:
            raise ValueError("either 'vocab_file' or 'pretrained_model' has to be specified")

        if hasattr(self.tokenizer, "vocab"):
            self.vocab_size = len(self.tokenizer.vocab)

        special_tokens = MODEL_SPECIAL_TOKENS[bert_derivative]
        for k, v in special_tokens.items():
            setattr(self, k, v)
        self.never_split = tuple(special_tokens.values())

    def text_to_tokens(self, text):
        tokens = self.tokenizer.tokenize(text)
        return tokens

    def tokens_to_text(self, tokens):
        text = self.tokenizer.convert_tokens_to_string(tokens)
        return remove_spaces(handle_quotes(text.strip()))

    def token_to_id(self, token):
        return self.tokens_to_ids([token])[0]

    def tokens_to_ids(self, tokens):
        ids = self.tokenizer.convert_tokens_to_ids(tokens)
        return ids

    def ids_to_tokens(self, ids):
        tokens = self.tokenizer.convert_ids_to_tokens(ids)
        return tokens

    def text_to_ids(self, text):
        tokens = self.text_to_tokens(text)
        ids = self.tokens_to_ids(tokens)
        return ids

    def ids_to_text(self, ids):
        tokens = self.ids_to_tokens(ids)
        tokens_clean = [t for t in tokens if t not in self.never_split]
        text = self.tokens_to_text(tokens_clean)
        return text

    def add_special_tokens(self, special_tokens_dict):
        # TODO: check that input's key is 'additional_special_tokens' or name of one of the previous special tokens
        # and add them to never split list
        self.tokenizer.add_special_tokens(special_tokens_dict)
        # self.never_split.append(special_tokens_dict["additional_special_tokens"])
        self.vocab_size = len(self.tokenizer)

    @property
    def pad_id(self):
        return self.tokens_to_ids([getattr(self, 'pad_token')])[0]

    @property
    def bos_id(self):
        return self.tokens_to_ids([getattr(self, 'bos_token')])[0]

    @property
    def eos_id(self):
        return self.tokens_to_ids([getattr(self, 'eos_token')])[0]

    @property
    def sep_id(self):
        return self.tokens_to_ids([getattr(self, 'sep_token')])[0]

    @property
    def cls_id(self):
        return self.tokens_to_ids([getattr(self, 'cls_token')])[0]



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2018 The Google AI Language Team Authors and
# The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import os

from nemo import logging
from nemo.collections.nlp.data.datasets.datasets_utils.datasets_processing import DataProcessor

__all__ = [
    'ColaProcessor',
    'MnliProcessor',
    'MnliMismatchedProcessor',
    'MrpcProcessor',
    'Sst2Processor',
    'StsbProcessor',
    'QqpProcessor',
    'QnliProcessor',
    'RteProcessor',
    'WnliProcessor',
]


class MrpcProcessor(DataProcessor):
    """Processor for the MRPC data set (GLUE version)."""

    def get_train_examples(self, data_dir):
        """See base class."""
        logging.info(f'LOOKING AT {os.path.join(data_dir, "train.tsv")}')
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "train.tsv")), "train")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "dev.tsv")), "dev")

    def get_labels(self):
        """See base class."""
        return ["0", "1"]

    def _create_examples(self, lines, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, line) in enumerate(lines):
            if i == 0:
                continue
            guid = "%s-%s" % (set_type, i)
            text_a = line[3]
            text_b = line[4]
            label = line[0]
            examples.append(InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
        return examples


class MnliProcessor(DataProcessor):
    """Processor for the MultiNLI data set (GLUE version)."""

    def get_train_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "train.tsv")), "train")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "dev_matched.tsv")), "dev_matched")

    def get_labels(self):
        """See base class."""
        return ["contradiction", "entailment", "neutral"]

    def _create_examples(self, lines, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, line) in enumerate(lines):
            if i == 0:
                continue
            guid = "%s-%s" % (set_type, line[0])
            text_a = line[8]
            text_b = line[9]
            label = line[-1]
            examples.append(InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
        return examples


class MnliMismatchedProcessor(MnliProcessor):
    """Processor for the MultiNLI Mismatched data set (GLUE version)."""

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "dev_mismatched.tsv")), "dev_matched")


class ColaProcessor(DataProcessor):
    """Processor for the CoLA data set (GLUE version)."""

    def get_train_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "train.tsv")), "train")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "dev.tsv")), "dev")

    def get_labels(self):
        """See base class."""
        return ["0", "1"]

    def _create_examples(self, lines, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, line) in enumerate(lines):
            guid = "%s-%s" % (set_type, i)
            text_a = line[3]
            label = line[1]
            examples.append(InputExample(guid=guid, text_a=text_a, text_b=None, label=label))
        return examples


class Sst2Processor(DataProcessor):
    """Processor for the SST-2 data set (GLUE version)."""

    def get_train_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "train.tsv")), "train")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "dev.tsv")), "dev")

    def get_labels(self):
        """See base class."""
        return ["0", "1"]

    def _create_examples(self, lines, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, line) in enumerate(lines):
            if i == 0:
                continue
            guid = "%s-%s" % (set_type, i)
            text_a = line[0]
            label = line[1]
            examples.append(InputExample(guid=guid, text_a=text_a, text_b=None, label=label))
        return examples


class StsbProcessor(DataProcessor):
    """Processor for the STS-B data set (GLUE version)."""

    def get_train_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "train.tsv")), "train")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "dev.tsv")), "dev")

    def get_labels(self):
        """See base class."""
        return [None]

    def _create_examples(self, lines, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, line) in enumerate(lines):
            if i == 0:
                continue
            guid = "%s-%s" % (set_type, line[0])
            text_a = line[7]
            text_b = line[8]
            label = line[-1]
            examples.append(InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
        return examples


class QqpProcessor(DataProcessor):
    """Processor for the QQP data set (GLUE version)."""

    def get_train_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "train.tsv")), "train")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "dev.tsv")), "dev")

    def get_labels(self):
        """See base class."""
        return ["0", "1"]

    def _create_examples(self, lines, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, line) in enumerate(lines):
            if i == 0:
                continue
            guid = "%s-%s" % (set_type, line[0])
            try:
                text_a = line[3]
                text_b = line[4]
                label = line[5]
            except IndexError:
                continue
            examples.append(InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
        return examples


class QnliProcessor(DataProcessor):
    """Processor for the QNLI data set (GLUE version)."""

    def get_train_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "train.tsv")), "train")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "dev.tsv")), "dev_matched")

    def get_labels(self):
        """See base class."""
        return ["entailment", "not_entailment"]

    def _create_examples(self, lines, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, line) in enumerate(lines):
            if i == 0:
                continue
            guid = "%s-%s" % (set_type, line[0])
            text_a = line[1]
            text_b = line[2]
            label = line[-1]
            examples.append(InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
        return examples


class RteProcessor(DataProcessor):
    """Processor for the RTE data set (GLUE version)."""

    def get_train_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "train.tsv")), "train")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "dev.tsv")), "dev")

    def get_labels(self):
        """See base class."""
        return ["entailment", "not_entailment"]

    def _create_examples(self, lines, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, line) in enumerate(lines):
            if i == 0:
                continue
            guid = "%s-%s" % (set_type, line[0])
            text_a = line[1]
            text_b = line[2]
            label = line[-1]
            examples.append(InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
        return examples


class WnliProcessor(DataProcessor):
    """Processor for the WNLI data set (GLUE version)."""

    def get_train_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "train.tsv")), "train")

    def get_dev_examples(self, data_dir):
        """See base class."""
        return self._create_examples(self._read_tsv(os.path.join(data_dir, "dev.tsv")), "dev")

    def get_labels(self):
        """See base class."""
        return ["0", "1"]

    def _create_examples(self, lines, set_type):
        """Creates examples for the training and dev sets."""
        examples = []
        for (i, line) in enumerate(lines):
            if i == 0:
                continue
            guid = "%s-%s" % (set_type, line[0])
            text_a = line[1]
            text_b = line[2]
            label = line[-1]
            examples.append(InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
        return examples


class InputExample(object):
    """A single training/test example for simple sequence classification."""

    def __init__(self, guid, text_a, text_b=None, label=None):
        """Constructs a InputExample.

        Args:
            guid: Unique id for the example.
            text_a: string. The untokenized text of the first sequence.
            For single sequence tasks, only this sequence must be specified.
            text_b: (Optional) string. The untokenized text of the second
            sequence. Only must be specified for sequence pair tasks.
            label: (Optional) string. The label of the example. This should be
            specified for train and dev examples, but not for test examples.
        """
        self.guid = guid
        self.text_a = text_a
        self.text_b = text_b
        self.label = label



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import os

from .package_info import (
    __contact_emails__,
    __contact_names__,
    __description__,
    __download_url__,
    __homepage__,
    __keywords__,
    __license__,
    __package_name__,
    __repository_url__,
    __shortversion__,
    __version__,
)

if "NEMO_PACKAGE_BUILDING" not in os.environ:
    from nemo.utils import logging, logging_mode

    from nemo import backends
    from nemo import core
    from nemo import utils
    from nemo import collections

    from nemo.backends.pytorch import tutorials



# -*- coding: utf-8 -*-

# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from collections.abc import MutableMapping
from typing import Any, Dict, List, Optional

from frozendict import frozendict

from nemo.utils import logging
from nemo.utils.neural_graph.connection import StepModulePort


class GraphOutput(object):
    """ A helper class represenging a single bound output. """

    def __init__(self, ntype: "NeuralType", producer_step_module_port: StepModulePort):
        """ 
        Initializes object.

        Args:
            ntype: a NeuralType object.
            producer_step_module_port: a producer StepModulePort tuple (step number (module name), port name).
        """
        self._ntype = ntype
        self._producer_step_module_port = producer_step_module_port

    @property
    def ntype(self) -> "NeuralType":
        """ 
            Returns:
                NeuralType of a given output.
        """
        return self._ntype

    @property
    def producer_step_module_port(self) -> StepModulePort:
        """ Returns producer step port (step number (module), port name) tuple. """
        return self._producer_step_module_port


class GraphOutputs(MutableMapping):
    '''
        A specialized dictionary that contains bound outputs of a Neural Graph.
        In fact stores two lists of "outputs":
            - "default" outputs with default keys taken from outputs of modules, and
            - "manual" used for specifying the subset of outputs.
        When accessing the outputs, it returns the one of those two lists following the rule:
        return "manual" outputs if they were define (at least one manual output defined by the user),
        otherwise return the "default" outputs.
    '''

    def __init__(self, tensors_ref):
        """
            Initializes two (empty) dictionaries. 

            Args:
                tensors_ref - reference to neural graph's tensor (dict of dict).
        """

        # Tensors[step][output_port_name] passed from the external neural graph object.
        self._tensors_ref = tensors_ref

        # This dictionary stores the bound outputs collected during the "default" recording of produced tensors.
        # As they are using the default port names, the second/next tensor published on the same port
        # will generate a new unique name following the (step_number.module.port_name) pattern.
        self._default_outputs = {}

        # This dictionary stores list of outputs of modules "manually" bound by the user.
        # In this case tring to overwriting the existing ports with new tensors will be forbidden (Exception).
        self._manual_outputs = {}

    def __setitem__(self, key: str, value: "NmTensor"):
        """
            This method is used to set the manual output - creates a GraphOutput item and adds it to the list.
            
            Args:
                key: The name of the output (port).
                value: NmTensor that will be used to create a given GraphOutput.
        """
        # Make sure that user passed a NmTensor.
        if type(value).__name__ != "NmTensor":
            raise TypeError("Port `{}` definition must be must be set using a NmTensor".format(key))

        if key in self._manual_outputs.keys():
            raise KeyError("Overwriting of a port `{}` that was previously manually bound is not allowed".format(key))

        # Ok, set thee "manual" output.
        self._manual_outputs[key] = GraphOutput(value.ntype, value.producer_step_module_port)

    def __getitem__(self, key: str) -> GraphOutput:
        """
            Returns the bound output associated with the given key.
            Uses default or manual dict depending whether there are some manual outputs or not.

            Args:
                key: Name of the bound input.
        """
        if len(self._manual_outputs) > 0:
            return self._manual_outputs[key]
        else:  # Use default dict.
            return self._default_outputs[key]

    def __delitem__(self, key: str):
        """
            Raises:
                TypeError as deletion of a bound input port is not allowed.
        """
        raise TypeError("Deleting a bound output is not allowed")

    def __iter__(self):
        """
            Returns:
                Iterator over the outputs - depending whether there are some manual outputs or not.
        """
        if len(self._manual_outputs) > 0:
            return iter(self._manual_outputs)
        else:  # Use default dict.
            return iter(self._default_outputs)

    def __len__(self) -> int:
        """
            Returns:
                The number of outputs - depending whether there are some manual outputs or not.
        """
        if len(self._manual_outputs) > 0:
            return len(self._manual_outputs)
        else:  # Use default dict.
            return len(self._default_outputs)

    def bind(self, tensors_ref: List["NmTensor"], port_names: Optional[str] = None):
        """
            Binds the "default" outputs.

            Args:
                tensors_ref: List of tensors to be added.
                port_names: List of port names (visible outside). If None: using internal tensor "output port names".
        """
        # Set names.
        if port_names is None:
            port_names = [tensor.name for tensor in tensors_ref]

        for name, tensor in zip(port_names, tensors_ref):
            # Check the presence of the port name in "default" dictionary.
            if name in self._default_outputs.keys():
                # Name present - use the name being combination of producer and port names.
                name = (
                    str(tensor.producer_step_number) + "_" + tensor.producer_name + "_" + tensor.name
                )  # last = port name

                logging.debug(
                    "Setting unique name of the default output port `{}` produced in step {} by `{}` to `{}`".format(
                        tensor.name, tensor.producer_step_number, tensor.producer_name, name
                    )
                )
            # Store the output.
            self._default_outputs[name] = GraphOutput(tensor.ntype, tensor.producer_step_module_port)

    @property
    def definitions(self) -> Dict[str, GraphOutput]:
        """
            Property returns definitions of the output ports by extracting them on the fly from the bound outputs.

            ..info:
                This property actually returns a FrozenDict containing port definitions to indicate that
                port definitions SHOULD not be used during the actual binding.
            

            Returns:
                Dictionary of neural types associated with bound outputs.
        """
        # Get the right output dictionary.
        d = self._manual_outputs if len(self._manual_outputs) > 0 else self._default_outputs

        # Extract port definitions (Neural Types) and return an immutable dictionary,
        # so the user won't be able to modify its content by an accident!
        return frozendict({k: v.ntype for k, v in d.items()})

    @property
    def tensors(self) -> Dict[str, "NmTensor"]:
        """
            Property returns output tensors by extracting them on the fly from the bound outputs.

            Returns:
                Dictionary of tensors in the format (output-name: tensor).
        """
        # Get the right output dictionary.
        d = self._manual_outputs if len(self._manual_outputs) > 0 else self._default_outputs

        output_tensors = {}
        # Get tensors by acessing the producer-ports.
        # At that point all keys (k) are unigue - we made sure of that during binding/__setitem__.
        for k, v in d.items():
            producer_step = v.producer_step_module_port.step_number
            producer_port_name = v.producer_step_module_port.port_name
            # Find the right output tensor.
            tensor = self._tensors_ref[producer_step][producer_port_name]
            # Add it to the dictionary.
            output_tensors[k] = tensor
        # Return the result as an immutable dictionary,
        # so the user won't be able to modify its content by an accident!
        return frozendict(output_tensors)

    @property
    def tensor_list(self) -> List["NmTensor"]:
        """
            Property returns output tensors by extracting them on the fly from the bound outputs.
            
            Returns:
                List of tensors.

        """
        # Get the right output dictionary.
        d = self._manual_outputs if len(self._manual_outputs) > 0 else self._default_outputs

        output_tensor_list = []
        # Get tensors by acessing the producer-ports.
        for k, v in d.items():
            producer_step = v.producer_step_module_port.step_number
            producer_port_name = v.producer_step_module_port.port_name
            # Find the right output tensor.
            tensor = self._tensors_ref[producer_step][producer_port_name]
            # Add it to the list.
            output_tensor_list.append(tensor)
        # Return the result.
        return output_tensor_list

    def serialize(self) -> Dict[str, Any]:
        """ Method responsible for serialization of the graph outputs.

            Returns:
                List containing mappings (step.module.output_port -> output | ntype).
        """
        serialized_outputs = {"mappings": []}

        # Get the right output dictionary.
        if len(self._manual_outputs) > 0:
            serialized_outputs["type"] = "manual"
            d = self._manual_outputs
        else:
            serialized_outputs["type"] = "default"
            d = self._default_outputs

        # Iterate through "bindings" (GraphOutputs).
        for key, binding in d.items():
            # Serialize: step.module.port -> output | ntype.
            smp = binding.producer_step_module_port
            source = str(smp.step_number) + "." + smp.module_name + "." + smp.port_name
            # Get type.
            ntype_str = str(binding.ntype)
            # Serialize!
            serialized_outputs["mappings"].append(source + "->" + key + " | " + ntype_str)
        # Return the result.
        return serialized_outputs

    def deserialize(self, serialized_outputs: Dict[str, Any], modules: Dict[str, 'NeuralModule']):
        """ 
            Method responsible for deserialization of graph outputs.

            Args:
                serialized_outputs: A list of serialized outputs in the form of ("step.module.output_port->key | ntype")
                modules: List of modules required for neural type copying/checking.
        """
        # Check type.
        if serialized_outputs["type"] == "default":
            # We still need to deserialize.
            # Use-case: deserialization of a graph with nested graph with bound output.
            d = self._default_outputs
        else:
            d = self._manual_outputs

        # Iterate through serialized inputs one by one.
        for i in serialized_outputs["mappings"]:
            # Deserialize!
            [producer, key_ntype] = i.split("->")
            [key, ntype_str] = key_ntype.split(" | ")
            [step_number, producer_name, producer_port_name] = producer.split(".")
            # Get neural type from module output port definition.
            ntype = modules[producer_name].output_ports[producer_port_name]

            # Make sure the graph bound port type matches the deserialized type.
            assert ntype_str == str(ntype)

            # Create a new input.
            go = GraphOutput(ntype, StepModulePort(int(step_number), producer_name, producer_port_name))
            d[key] = go

        # Done.



# noinspection PyPep8Naming

from torch import nn

from nemo.backends.pytorch.nm import TrainableNM
from nemo.collections.asr.jasper import init_weights as jasper_init_weights
from nemo.core.neural_types import *
from nemo.utils.decorators import add_port_docs


class JasperRNNConnector(TrainableNM):
    """Connector between jasper encoder and some other module, that does
    change number of channels.

    Args:
        in_channels: Number of channels of input tensor
        out_channels: Number of channels to reshape to

    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        # return {'tensor': NeuralType({0: AxisType(BatchTag), 1: AxisType(ChannelTag), 2: AxisType(TimeTag),})}
        return {'tensor': NeuralType(('B', 'D', 'T'), ChannelType())}

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        tensor:
            0: AxisType(BatchTag)

            1: AxisType(TimeTag)

            2: AxisType(ChannelTag)
        """
        # return {'tensor': NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag), 2: AxisType(ChannelTag),})}
        return {'tensor': NeuralType(('B', 'T', 'D'), ChannelType())}

    def __init__(self, in_channels, out_channels):
        super().__init__()

        self.icnn = nn.Conv1d(in_channels, out_channels, kernel_size=1, bias=True)
        self.bn = nn.BatchNorm1d(out_channels)

        self.apply(jasper_init_weights)
        self.to(self._device)

    def forward(self, tensor):
        # tensor = F.relu(tensor)
        # tensor = F.dropout(tensor, 0.2)
        tensor = self.icnn(tensor)
        tensor = self.bn(tensor)
        tensor = tensor.transpose(1, 2)
        return tensor



# Copyright (c) 2019 NVIDIA Corporation
import torch

from .parts.datasets import AudioOnlyDataset
from nemo.backends.pytorch.nm import DataLayerNM
from nemo.core import DeviceType
from nemo.core.neural_types import AudioSignal, LengthsType, NeuralType
from nemo.utils import logging
from nemo.utils.decorators import add_port_docs


class AudioDataLayer(DataLayerNM):
    """
    Data Layer for general speech tasks that loads only the audio.

    Module which reads speech data. It accepts comma-separated
    JSON manifest files describing the wav audio files and their metadata.
    JSON files should be of the following format::

        {"audio_filepath": path_to_wav_0, "duration": time_in_sec_0}
        ...
        {"audio_filepath": path_to_wav_n, "duration": time_in_sec_n}


    Args:
        manifest_filepath (str): path to JSON containing data.
        batch_size (int): batch sizelse.
        min_duration (float): All training files which have a duration less
            than min_duration are dropped. Note: Duration is read from the
            manifest JSON.
            Defaults to 0.1.
        max_duration (float): All training files which have a duration more
            than max_duration are dropped. Note: Duration is read from the
            manifest JSON.
            Defaults to None.
        trim_silence (bool): Whether to use trim silence from beginning and end
            of audio signal using librosa.effects.trim().
            Defaults to False.
        drop_last (bool): See PyTorch DataLoader.
            Defaults to False.
        shuffle (bool): See PyTorch DataLoader.
            Defaults to True.
        num_workers (int): See PyTorch DataLoader.
            Defaults to 0.
        n_segments (int): Number of samples to load per audiofile.
            Defaults to 0 which indicates to load the whole file.
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "audio_signal": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag)}),
            # "a_sig_length": NeuralType({0: AxisType(BatchTag)}),
            "audio_signal": NeuralType(('B', 'T'), AudioSignal(freq=self.sample_rate)),
            "a_sig_length": NeuralType(tuple('B'), LengthsType()),
        }

    def __init__(
        self,
        manifest_filepath,
        batch_size,
        sample_rate,
        min_duration=0.1,
        max_duration=None,
        trim_silence=False,
        drop_last=False,
        shuffle=True,
        num_workers=0,
        n_segments=0,
    ):
        super().__init__()
        self.sample_rate = sample_rate

        self._dataset = AudioOnlyDataset(
            manifest_filepath=manifest_filepath,
            max_duration=max_duration,
            min_duration=min_duration,
            trim=trim_silence,
            n_segments=n_segments,
        )

        sampler = None
        if self._placement == DeviceType.AllGpu:
            logging.info('Parallelizing DATALAYER')
            sampler = torch.utils.data.distributed.DistributedSampler(self._dataset)

        self._dataloader = torch.utils.data.DataLoader(
            dataset=self._dataset,
            batch_size=batch_size,
            collate_fn=self._dataset.AudioCollateFunc,
            drop_last=drop_last,
            shuffle=shuffle if sampler is None else False,
            sampler=sampler,
            num_workers=num_workers,
        )

    def __len__(self):
        return len(self._dataset)

    @property
    def dataset(self):
        return None

    @property
    def data_iterator(self):
        return self._dataloader



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2018 The Google AI Language Team Authors and
# The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from typing import List, Optional

from transformers import BERT_PRETRAINED_CONFIG_ARCHIVE_MAP, BERT_PRETRAINED_MODEL_ARCHIVE_LIST, BertConfig, BertModel

from nemo.backends.pytorch.nm import TrainableNM
from nemo.core.neural_modules import PretrainedModelInfo
from nemo.core.neural_types import ChannelType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['BERT']


class BERT(TrainableNM):
    """
    BERT wraps around the Huggingface implementation of BERT from their
    transformers repository for easy use within NeMo.

    Args:
        pretrained_model_name (str): If using a pretrained model, this should
            be the model's name. Otherwise, should be left as None.
        config_filename (str): path to model configuration file. Optional.
        vocab_size (int): Size of the vocabulary file, if not using a
            pretrained model.
        hidden_size (int): Size of the encoder and pooler layers.
        num_hidden_layers (int): Number of hidden layers in the encoder.
        num_attention_heads (int): Number of attention heads for each layer.
        intermediate_size (int): Size of intermediate layers in the encoder.
        hidden_act (str): Activation function for encoder and pooler layers;
            "gelu", "relu", and "swish" are supported.
        max_position_embeddings (int): The maximum number of tokens in a
        sequence.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        input_ids: input token ids
        token_type_ids: segment type ids
        attention_mask: attention mask
        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "token_type_ids": NeuralType(('B', 'T'), ChannelType()),
            "attention_mask": NeuralType(('B', 'T'), ChannelType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        hidden_states: output embedding 
        """
        return {"hidden_states": NeuralType(('B', 'T', 'D'), ChannelType())}

    def __init__(
        self,
        pretrained_model_name=None,
        config_filename=None,
        vocab_size=None,
        hidden_size=768,
        num_hidden_layers=12,
        num_attention_heads=12,
        intermediate_size=3072,
        hidden_act="gelu",
        max_position_embeddings=512,
        hidden_dropout_prob=0.1,
        attention_probs_dropout_prob=0.1,
        type_vocab_size=2,
        initializer_range=0.02,
        layer_norm_eps=1e-12,
    ):

        super().__init__()

        # Check that only one of pretrained_model_name, config_filename, and
        # vocab_size was passed in
        total = 0
        if pretrained_model_name is not None:
            total += 1
        if config_filename is not None:
            total += 1
        if vocab_size is not None:
            total += 1

        if total != 1:
            raise ValueError(
                "Only one of pretrained_model_name, vocab_size, "
                + "or config_filename should be passed into the "
                + "BERT constructor."
            )

        # TK: The following code checks the same once again.
        if vocab_size is not None:
            config = BertConfig(
                vocab_size_or_config_json_file=vocab_size,
                vocab_size=vocab_size,
                hidden_size=hidden_size,
                num_hidden_layers=num_hidden_layers,
                num_attention_heads=num_attention_heads,
                intermediate_size=intermediate_size,
                hidden_act=hidden_act,
                max_position_embeddings=max_position_embeddings,
            )
            model = BertModel(config)
        elif pretrained_model_name is not None:
            model = BertModel.from_pretrained(pretrained_model_name)
        elif config_filename is not None:
            config = BertConfig.from_json_file(config_filename)
            model = BertModel(config)
        else:
            raise ValueError(
                "Either pretrained_model_name or vocab_size must" + " be passed into the BERT constructor"
            )

        model.to(self._device)

        self.add_module("bert", model)
        self.config = model.config
        self._hidden_size = model.config.hidden_size

    def resize_token_embeddings(self, new_vocab_size):
        """ Resize input token embeddings matrix of the model if new_num_tokens != config.vocab_size.
        Take care of tying weights embeddings afterwards if the model class has a `tie_weights()` method.

        Args:
            new_vocab_size: (`optional`) int:
                New number of tokens in the embedding matrix. Increasing the size will add newly initialized 
                vectors at the end. Reducing the size will remove vectors from the end.
                If not provided or None: does nothing and just returns a pointer to the input tokens 
                ``torch.nn.Embeddings`` Module of the model.

        Return: ``torch.nn.Embeddings``
            Pointer to the input tokens Embeddings Module of the model
        """
        return self.bert.resize_token_embeddings(new_vocab_size)

    @property
    def hidden_size(self):
        """
            Property returning hidden size.

            Returns:
                Hidden size.
        """
        return self._hidden_size

    @staticmethod
    def list_pretrained_models() -> Optional[List[PretrainedModelInfo]]:
        pretrained_models = []
        for key in BERT_PRETRAINED_MODEL_ARCHIVE_LIST:
            model_info = PretrainedModelInfo(
                pretrained_model_name=key,
                description="weights by HuggingFace",
                parameters=BERT_PRETRAINED_CONFIG_ARCHIVE_MAP[key],
                location="",
            )
            pretrained_models.append(model_info)
        return pretrained_models

    def forward(self, input_ids, token_type_ids, attention_mask):
        return self.bert(input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask)[0]



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from enum import Enum
from typing import List

import numpy as np
import torch

from nemo import logging
from nemo.backends.pytorch.nm import DataLayerNM
from nemo.core.neural_types import *

__all__ = ['MultiDataLayer', 'DataCombination']


class DataCombination(Enum):
    CROSSPRODUCT = 1
    ZIP = 2


class MultiDataLayer(DataLayerNM):
    def __init__(
        self,
        data_layers: List[DataLayerNM],
        batch_size: int,
        shuffle: bool = False,
        combination_mode: DataCombination = DataCombination.CROSSPRODUCT,
        port_names: List[str] = None,
    ):
        """
        data_layers: (list) of DataLayerNM objects
        batch_size: (int) batchsize when the underlying dataset is loaded
        combination_mode: (DataCombination) defines how to combine the datasets.
        shuffle: (bool) whether underlying multi dataset should be shuffled in each epoch
        port_names: List(str) user can override all port names if specified 
        """
        super().__init__()
        self._data_layers = data_layers
        self._batch_size = batch_size
        self._shuffle = shuffle
        self._combination_mode = combination_mode
        self._port_names = port_names
        self._dataset = MultiDataset(
            datasets=[dl.dataset for dl in self._data_layers], combination_mode=combination_mode
        )

        self._ports = dict()
        if self._port_names:
            i = 0
            for dl in self._data_layers:
                for _, port_type in dl.output_ports.items():
                    self._ports[self._port_names[i]] = port_type
                    i += 1
        else:
            for dl_idx, dl in enumerate(self._data_layers):
                for port_name, port_type in dl.output_ports.items():
                    if port_name in self._ports:
                        logging.warning(f"name collision {port_name}, will rename")
                        self._ports[f"{port_name}_{dl_idx}"] = port_type
                    else:
                        self._ports[port_name] = port_type

    @property
    def output_ports(self):
        """Return: dict
        Returns union of all individual data_layer output ports
        In case of name collision, resolve by renaming 
        """
        return self._ports

    def __len__(self):
        return len(self._dataset)

    @property
    def dataset(self):
        return self._dataset

    @property
    def data_iterator(self):
        return None


class MultiDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        datasets: List[torch.utils.data.Dataset],
        combination_mode: DataCombination = DataCombination.CROSSPRODUCT,
    ):
        """
        Datasets: list of torch.utils.data.Dataset objects.
        combination_mode: DataCombination, defines how to combine the datasets, Options are [DataCombination.CROSSPRODUCT, DataCombination.ZIP]. 
        """
        self.datasets = datasets
        self.combination_mode = combination_mode
        if self.combination_mode == DataCombination.CROSSPRODUCT:
            self.len = np.prod([len(d) for d in self.datasets])
        elif self.combination_mode == DataCombination.ZIP:
            ds_lens = [len(d) for d in self.datasets]
            self.len = np.min(ds_lens)
            if len(set(ds_lens)) != 1:
                raise ValueError("datasets do not have equal lengths.")
        else:
            raise ValueError("combination_mode unknown")

    def __getitem__(self, i):
        """
        Returns list [x1, x2, ...xn] where x1 \in D1, x2 \in D2, ..., xn \in Dn
        """

        return [x for d in self.datasets for x in d[i % len(d)]]

    def __len__(self):
        """
        Returns length of this dataset (int).
        In case of  DataCombination.CROSSPRODUCT this would be prod(len(d) for d in self.datasets). 
        In case of  DataCombination.ZIP this would be min(len(d) for d in self.datasets) given that all datasets have same length. 
        """
        return self.len



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from torch import nn

from nemo.backends.pytorch.nm import TrainableNM
from nemo.core.neural_types import *
from nemo.utils.decorators import add_port_docs

__all__ = ['EncoderRNN']


class EncoderRNN(TrainableNM):
    """ Simple RNN-based encoder using GRU cells - with input/output port definitions used in TRADE. """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            'inputs': NeuralType(('B', 'T'), TokenIndex()),
            'input_lens': NeuralType(tuple('B'), Length()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            'outputs': NeuralType(('B', 'T', 'D'), ChannelType()),
            'hidden': NeuralType(('B', 'T', 'D'), ChannelType()),
        }

    def __init__(
        self, input_dim, emb_dim, hid_dim, dropout, n_layers=1, pad_idx=1, embedding_to_load=None, sum_hidden=True
    ):
        super().__init__()
        self.dropout = nn.Dropout(dropout)
        self.embedding = nn.Embedding(input_dim, emb_dim, padding_idx=pad_idx)
        if embedding_to_load is not None:
            self.embedding.weight.data.copy_(embedding_to_load)
        else:
            self.embedding.weight.data.normal_(0, 0.1)
        self.rnn = nn.GRU(emb_dim, hid_dim, n_layers, batch_first=True, dropout=dropout, bidirectional=True)
        self.sum_hidden = sum_hidden
        self.to(self._device)

    def forward(self, inputs, input_lens=None):
        embedded = self.embedding(inputs)
        embedded = self.dropout(embedded)
        if input_lens is not None:
            embedded = nn.utils.rnn.pack_padded_sequence(embedded, input_lens, batch_first=True)

        outputs, hidden = self.rnn(embedded)
        # outputs of shape (seq_len, batch, num_directions * hidden_size)
        # hidden of shape (num_layers * num_directions, batch, hidden_size)
        if input_lens is not None:
            outputs, _ = nn.utils.rnn.pad_packed_sequence(outputs, batch_first=True)
        else:
            outputs = outputs.transpose(0, 1)
        # outputs of shape: (batch, seq_len, num_directions * hidden_size)

        batch_size = hidden.size()[1]

        # separate final hidden states by layer and direction
        hidden = hidden.view(self.rnn.num_layers, 2 if self.rnn.bidirectional else 1, batch_size, self.rnn.hidden_size)
        hidden = hidden.transpose(2, 0).transpose(1, 2)
        # hidden shape: batch x num_layer x num_directions x hidden_size
        if self.sum_hidden and self.rnn.bidirectional:
            hidden = hidden[:, :, 0, :] + hidden[:, :, 1, :]
            outputs = outputs[:, :, : self.rnn.hidden_size] + outputs[:, :, self.rnn.hidden_size :]
        else:
            hidden = hidden.reshape(batch_size, self.rnn.num_layers, -1)
        # hidden is now of shape (batch, num_layer, [num_directions] * hidden_size)

        return outputs, hidden
